"""
    verktyg.routing
    ~~~~~~~~~~~~~~~

    When it comes to combining multiple controller or view functions (however
    you want to call them) you need a dispatcher.  A simple way would be
    applying regular expression tests on the ``PATH_INFO`` and calling
    registered callback functions that return the value then.

    This module implements a much more powerful system than simple regular
    expression matching because it can also convert values in the URLs and
    build URLs.

    Here a simple example that creates an URLMap for an application with
    two subdomains (www and kb) and some URL rules:

    >>> m = URLMap([
    ...     # Static URLs
    ...     Route('/', endpoint='static/index'),
    ...     Route('/about', endpoint='static/about'),
    ...     Route('/help', endpoint='static/help'),
    ...     # Knowledge Base
    ...     Subdomain('kb', [
    ...         Route('/', endpoint='kb/index'),
    ...         Route('/browse/', endpoint='kb/browse'),
    ...         Route('/browse/<int:id>/', endpoint='kb/browse'),
    ...         Route('/browse/<int:id>/<int:page>', endpoint='kb/browse')
    ...     ])
    ... ], default_subdomain='www')

    If the application doesn't use subdomains it's perfectly fine to not set
    the default subdomain and not use the `Subdomain` route factory.  The
    endpoint in the routes can be anything, for example import paths or unique
    identifiers.  The WSGI application can use those endpoints to get the
    handler for that URL.  It doesn't have to be a string at all but it's
    recommended.

    Now it's possible to create a URL adapter for one of the subdomains and
    build URLs:

    >>> c = m.bind('example.com')
    >>> c.build("kb/browse", dict(id=42))
    'http://kb.example.com/browse/42/'
    >>> c.build("kb/browse", dict())
    'http://kb.example.com/browse/'
    >>> c.build("kb/browse", dict(id=42, page=3))
    'http://kb.example.com/browse/42/3'
    >>> c.build("static/about")
    '/about'
    >>> c.build("static/index", force_external=True)
    'http://www.example.com/'

    >>> c = m.bind('example.com', subdomain='kb')
    >>> c.build("static/about")
    'http://www.example.com/about'

    The first argument to bind is the server name *without* the subdomain.
    Per default it will assume that the script is mounted on the root, but
    often that's not the case so you can provide the real mount point as
    second argument:

    >>> c = m.bind('example.com', '/applications/example')

    The third argument can be the subdomain, if not given the default
    subdomain is used.  For more details about binding have a look at the
    documentation of the `MapAdapter`.

    And here is how you can match URLs:

    >>> c = m.bind('example.com')
    >>> c.match("/")
    ('static/index', {})
    >>> c.match("/about")
    ('static/about', {})
    >>> c = m.bind('example.com', '/', 'kb')
    >>> c.match("/")
    ('kb/index', {})
    >>> c.match("/browse/42/23")
    ('kb/browse', {'id': 42, 'page': 23})

    If matching fails you get a `NotFound` exception, if the route thinks
    it's a good idea to redirect (for example because the URL was defined
    to have a slash at the end but the request was missing that slash) it
    will raise a `RequestRedirect` exception.  Both are subclasses of the
    `HTTPException` so you can use those errors as responses in the
    application.


    :copyright:
        (c) 2014 Ben Mather, based on Werkzeug, see AUTHORS for more details.
    :license:
        BSD, see LICENSE for more details.
"""
import re
import uuid
import posixpath
from urllib.parse import (
    urlencode, quote as urlquote, urljoin,
    urlunparse, ParseResult
)

from pprint import pformat

from verktyg.datastructures import ImmutableDict, MultiDict
from verktyg.wsgi import wsgi_decoding_dance
from verktyg.exceptions import HTTPException, NotFound


_route_re = re.compile(r'''
    (?P<static>[^<]*)                           # static route data
    <
    (?:
        (?P<converter>[a-zA-Z_][a-zA-Z0-9_]*)   # converter name
        (?:\((?P<args>.*?)\))?                  # converter arguments
        \:                                      # variable delimiter
    )?
    (?P<variable>[a-zA-Z_][a-zA-Z0-9_]*)        # variable name
    >
''', re.VERBOSE)
_simple_route_re = re.compile(r'<([^>]+)>')
_converter_args_re = re.compile(r'''
    ((?P<name>\w+)\s*=\s*)?
    (?P<value>
        True|False|
        \d+.\d+|
        \d+.|
        \d+|
        \w+|
        [urUR]?(?P<stringval>"[^"]*?"|'[^']*')
    )\s*,
''', re.VERBOSE | re.UNICODE)


_PYTHON_CONSTANTS = {
    'None': None,
    'True': True,
    'False': False
}


def _pythonize(value):
    if value in _PYTHON_CONSTANTS:
        return _PYTHON_CONSTANTS[value]
    for convert in int, float:
        try:
            return convert(value)
        except ValueError:
            pass
    if value[:1] == value[-1:] and value[0] in '"\'':
        value = value[1:-1]
    return str(value)


def parse_converter_args(argstr):
    argstr += ','
    args = []
    kwargs = {}

    for item in _converter_args_re.finditer(argstr):
        value = item.group('stringval')
        if value is None:
            value = item.group('value')
        value = _pythonize(value)
        if not item.group('name'):
            args.append(value)
        else:
            name = item.group('name')
            kwargs[name] = value

    return tuple(args), kwargs


def parse_route(route):
    """Parse a route and return it as generator. Each iteration yields tuples
    in the form ``(converter, arguments, variable)``. If the converter is
    `None` it's a static url part, otherwise it's a dynamic one.

    :internal:
    """
    pos = 0
    end = len(route)
    do_match = _route_re.match
    used_names = set()
    while pos < end:
        m = do_match(route, pos)
        if m is None:
            break
        data = m.groupdict()
        if data['static']:
            yield None, None, data['static']
        variable = data['variable']
        converter = data['converter'] or 'default'
        if variable in used_names:
            raise ValueError('variable name %r used twice.' % variable)
        used_names.add(variable)
        yield converter, data['args'] or None, variable
        pos = m.end()
    if pos < end:
        remaining = route[pos:]
        if '>' in remaining or '<' in remaining:
            raise ValueError('malformed url rule: %r' % route)
        yield None, None, remaining


class RoutingException(Exception):
    """Special exceptions that require the application to redirect, notifying
    about missing urls, etc.

    :internal:
    """


class RequestRedirect(HTTPException, RoutingException):
    """Raise if the router requests a redirect. This is for example the case if
    `strict_slashes` are activated and an url that requires a trailing slash.

    The attribute `new_url` contains the absolute destination url.
    """
    code = 301

    def __init__(self, new_url):
        RoutingException.__init__(self, new_url)
        self.new_url = new_url


class RequestSlash(RoutingException):
    """Internal exception."""


class RequestAliasRedirect(RoutingException):
    """This route is an alias and wants to redirect to the canonical URL."""

    def __init__(self, matched_values):
        self.matched_values = matched_values


class BuildError(RoutingException, LookupError):
    """Raised if the build system cannot find a URL for an endpoint with the
    values provided.
    """

    def __init__(self, endpoint, values):
        LookupError.__init__(self, endpoint, values)
        self.endpoint = endpoint
        self.values = values


class ValidationError(ValueError):
    """Validation error.  If a route converter raises this exception the route
    does not match the current URL and the next URL is tried.
    """


class RouteFactory(object):
    """As soon as you have more complex URL setups it's a good idea to use
    route factories to avoid repetitive tasks.  Some of them are builtin,
    others can be added by subclassing `RouteFactory` and overriding
    `get_routes`.
    """

    def get_routes(self, router):
        """Subclasses of `RouteFactory` have to override this method and return
        an iterable of routes."""
        raise NotImplementedError()


class Subdomain(RouteFactory):
    """All URLs provided by this factory have the subdomain set to a
    specific domain. For example if you want to use the subdomain for
    the current language this can be a good setup::

        router = URLMap([
            Route('/', endpoint='#select_language'),
            Subdomain('<string(length=2):lang_code>', [
                Route('/', endpoint='index'),
                Route('/about', endpoint='about'),
                Route('/help', endpoint='help')
            ])
        ])

    All the rules except for the ``'#select_language'`` endpoint will now
    listen on a two letter long subdomain that holds the language code
    for the current request.
    """

    def __init__(self, subdomain, routes):
        self.subdomain = subdomain
        self.routes = routes

    def get_routes(self, router):
        for routefactory in self.routes:
            for route in routefactory.get_routes(router):
                route = route.empty()
                route.subdomain = self.subdomain
                yield route


class Submount(RouteFactory):
    """Like `Subdomain` but prefixes the URL rule with a given string::

        router = URLMap([
            Route('/', endpoint='index'),
            Submount('/blog', [
                Route('/', endpoint='blog/index'),
                Route('/entry/<entry_slug>', endpoint='blog/show')
            ])
        ])

    Now the route ``'blog/show'`` matches ``/blog/entry/<entry_slug>``.
    """

    def __init__(self, path, routes):
        self.path = path.rstrip('/')
        self.routes = routes

    def get_routes(self, router):
        for routefactory in self.routes:
            for route in routefactory.get_routes(router):
                route = route.empty()
                route.route = self.path + route.route
                yield route


class EndpointPrefix(RouteFactory):
    """Prefixes all endpoints (which must be strings for this factory) with
    another string. This can be useful for sub applications::

        router = URLMap([
            Route('/', endpoint='index'),
            EndpointPrefix('blog/', [Submount('/blog', [
                Route('/', endpoint='index'),
                Route('/entry/<entry_slug>', endpoint='show')
            ])])
        ])
    """

    def __init__(self, prefix, routes):
        self.prefix = prefix
        self.routes = routes

    def get_routes(self, router):
        for routefactory in self.routes:
            for route in routefactory.get_routes(router):
                route = route.empty()
                route.endpoint = self.prefix + route.endpoint
                yield route


class RouteTemplate(object):
    """Returns copies of the routes wrapped and expands string templates in
    the endpoint, route, defaults or subdomain sections.

    Here a small example for such a route template::

        from verktyg.routing import URLMap, Route, RouteTemplate

        resource = RouteTemplate([
            Route('/{name}/', endpoint='{name}.list'),
            Route('/{name}/<int:id>', endpoint='{name}.show')
        ])

        router = URLMap([resource(name='user'), resource(name='page')])

    When a route template is called the keyword arguments are used to
    replace the placeholders in all the string parameters.
    """

    def __init__(self, routes):
        self.routes = list(routes)

    def __call__(self, *args, **kwargs):
        return RouteTemplateFactory(self.routes, dict(*args, **kwargs))


class RouteTemplateFactory(RouteFactory):
    """A factory that fills in template variables into routes.  Used by
    `RouteTemplate` internally.

    :internal:
    """

    def __init__(self, routes, context):
        self.routes = routes
        self.context = context

    def get_routes(self, router):
        for routefactory in self.routes:
            for route in routefactory.get_routes(router):
                new_defaults = subdomain = None
                if route.defaults:
                    new_defaults = {}
                    for key, value in route.defaults.items():
                        if isinstance(value, str):
                            value = value.format(**self.context)
                        new_defaults[key] = value
                if route.subdomain is not None:
                    subdomain = route.subdomain.format(**self.context)
                new_endpoint = route.endpoint
                if isinstance(new_endpoint, str):
                    new_endpoint = new_endpoint.format(**self.context)
                yield Route(
                    route.route.format(**self.context),
                    new_defaults,
                    subdomain,
                    route.build_only,
                    new_endpoint,
                    route.strict_slashes
                )


class Route(RouteFactory):
    """A Route represents one URL pattern.  There are some options for `Route`
    that change the way it behaves and are passed to the `Route` constructor.
    Note that besides the route-string all arguments *must* be keyword
    arguments in order to not break the application on Verktyg upgrades.

    `string`
        Route strings basically are just normal URL paths with placeholders in
        the format ``<converter(arguments):name>`` where the converter and the
        arguments are optional.  If no converter is defined the `default`
        converter is used which means `string` in the normal configuration.

        URL routes that end with a slash are branch URLs, others are leaves.
        If you have `strict_slashes` enabled (which is the default), all
        branch URLs that are matched without a trailing slash will trigger a
        redirect to the same URL with the missing slash appended.

        The converters are defined on the `URLMap`.

    `endpoint`
        The endpoint for this route. This can be anything. A reference to a
        function, a string, a number etc.  The preferred way is using a string
        because the endpoint is used for URL generation.

    `defaults`
        An optional dict with defaults for other routes with the same endpoint.
        This is a bit tricky but useful if you want to have unique URLs::

            router = URLMap([
                Route('/all/', defaults={'page': 1}, endpoint='all_entries'),
                Route('/all/page/<int:page>', endpoint='all_entries')
            ])

        If a user now visits ``http://example.com/all/page/1`` he will be
        redirected to ``http://example.com/all/``.  If `redirect_defaults` is
        disabled on the `URLMap` instance this will only affect the URL
        generation.

    `subdomain`
        The subdomain route string for this route. If not specified the route
        only matches for the `default_subdomain` of the router.  If the router
        is not bound to a subdomain this feature is disabled.

        Can be useful if you want to have user profiles on different subdomains
        and all subdomains are forwarded to your application::

            router = URLMap([
                Route('/', subdomain='<username>', endpoint='user/homepage'),
                Route('/stats', subdomain='<username>', endpoint='user/stats')
            ])

    `strict_slashes`
        Override the `URLMap` setting for `strict_slashes` only for this route.
        If not specified the `URLMap` setting is used.

    `build_only`
        Set this to True and the route will never match but will create a URL
        that can be build. This is useful if you have resources on a subdomain
        or folder that are not handled by the WSGI application (like static
        data)

    `redirect_to`
        If given this must be either a string or callable.  In case of a
        callable it's called with the url adapter that triggered the match and
        the values of the URL as keyword arguments and has to return the target
        for the redirect, otherwise it has to be a string with placeholders in
        route syntax::

            def foo_with_slug(adapter, id):
                # ask the database for the slug for the old id.  this of
                # course has nothing to do with verktyg.
                return 'foo/' + Foo.get_slug_for_id(id)

            router = URLMap([
                Route('/foo/<slug>', endpoint='foo'),
                Route('/some/old/url/<slug>', redirect_to='foo/<slug>'),
                Route('/other/old/url/<int:id>', redirect_to=foo_with_slug)
            ])

        When the route is matched the routing system will raise a
        `RequestRedirect` exception with the target for the redirect.

        Keep in mind that the URL will be joined against the URL root of the
        script so don't use a leading slash on the target URL unless you
        really mean root of that domain.

    `alias`
        If enabled this route serves as an alias for another route with the
        same endpoint and arguments.

    `host`
        If provided and the router has host matching enabled this can be
        used to provide a match route for the whole host.  This also means
        that the subdomain feature is disabled.
    """

    def __init__(
        self, string, defaults=None, subdomain=None,
        build_only=False, endpoint=None, strict_slashes=None,
        redirect_to=None, alias=False, host=None,
    ):
        if not string.startswith('/'):
            raise ValueError('urls must start with a leading slash')
        self.route = string
        self.is_leaf = not string.endswith('/')

        self.router = None
        self.strict_slashes = strict_slashes
        self.subdomain = subdomain
        self.host = host
        self.defaults = defaults
        self.build_only = build_only
        self.alias = alias
        self.endpoint = endpoint
        self.redirect_to = redirect_to

        if defaults:
            self.arguments = {str(default) for default in defaults}
        else:
            self.arguments = set()
        self._trace = self._converters = self._regex = self._weights = None

    def empty(self):
        """Return an unbound copy of this route.  This can be useful if you
        want to reuse an already bound URL for another router."""
        defaults = None
        if self.defaults:
            defaults = dict(self.defaults)
        return Route(
            self.route, defaults, self.subdomain,
            self.build_only, self.endpoint, self.strict_slashes,
            self.redirect_to, self.alias, self.host,
        )

    def get_routes(self, router):
        yield self

    def refresh(self):
        """Rebinds and refreshes the URL.  Call this if you modified the
        route in place.

        :internal:
        """
        self.bind(self.router, rebind=True)

    def bind(self, router, rebind=False):
        """Bind the url to a router and create a regular expression based on
        the information from the route itself and the defaults from the router.

        :internal:
        """
        if self.router is not None and not rebind:
            raise RuntimeError((
                'url route {route!r} already bound to router {router!r}'
            ).format(route=self, router=self.router))
        self.router = router
        if self.strict_slashes is None:
            self.strict_slashes = router.strict_slashes
        if self.subdomain is None:
            self.subdomain = router.default_subdomain
        self.compile()

    def get_converter(self, variable_name, converter_name, args, kwargs):
        """Looks up the converter for the given parameter.
        """
        if converter_name not in self.router.converters:
            raise LookupError(
                'the converter %r does not exist' % converter_name
            )
        converter = self.router.converters[converter_name]
        return converter(self.router, *args, **kwargs)

    def compile(self):
        """Compiles the regular expression and stores it."""
        assert self.router is not None, 'route not bound'

        if self.router.host_matching:
            domain_route = self.host or ''
        else:
            domain_route = self.subdomain or ''

        self._trace = []
        self._converters = {}
        self._weights = []
        regex_parts = []

        def _build_regex(route):
            for converter, arguments, variable in parse_route(route):
                if converter is None:
                    regex_parts.append(re.escape(variable))
                    self._trace.append((False, variable))
                    for part in variable.split('/'):
                        if part:
                            self._weights.append((0, -len(part)))
                else:
                    if arguments:
                        c_args, c_kwargs = parse_converter_args(arguments)
                    else:
                        c_args = ()
                        c_kwargs = {}
                    convobj = self.get_converter(
                        variable, converter, c_args, c_kwargs)
                    regex_parts.append(
                        '(?P<%s>%s)' % (variable, convobj.regex)
                    )
                    self._converters[variable] = convobj
                    self._trace.append((True, variable))
                    self._weights.append((1, convobj.weight))
                    self.arguments.add(str(variable))

        _build_regex(domain_route)
        regex_parts.append('\\|')
        self._trace.append((False, '|'))
        _build_regex(self.is_leaf and self.route or self.route.rstrip('/'))
        if not self.is_leaf:
            self._trace.append((False, '/'))

        if self.build_only:
            return
        regex = r'^%s%s$' % (
            u''.join(regex_parts),
            '(?<!/)(?P<__suffix__>/?)'
            if (not self.is_leaf or not self.strict_slashes)
            else ''
        )
        self._regex = re.compile(regex, re.UNICODE)

    def match(self, path):
        """Check if the route matches a given path. Path is a string in the
        form ``"subdomain|/path"`` and is assembled by the router.  If
        the router is doing host matching the subdomain part will be the host
        instead.

        If the route matches a dict with the converted values is returned,
        otherwise the return value is `None`.

        :internal:
        """
        if not self.build_only:
            m = self._regex.search(path)
            if m is not None:
                groups = m.groupdict()
                # we have a folder like part of the url without a trailing
                # slash and strict slashes enabled. raise an exception that
                # tells the router to redirect to the same url but with a
                # trailing slash
                if (
                    self.strict_slashes and not self.is_leaf and
                    not groups.pop('__suffix__')
                ):
                    raise RequestSlash()
                # if we are not in strict slashes mode we have to remove
                # a __suffix__
                elif not self.strict_slashes:
                    del groups['__suffix__']

                result = {}
                for name, value in groups.items():
                    try:
                        value = self._converters[name].to_python(value)
                    except ValidationError:
                        return
                    result[str(name)] = value
                if self.defaults:
                    result.update(self.defaults)

                if self.alias and self.router.redirect_defaults:
                    raise RequestAliasRedirect(result)

                return result

    def build(self, values, append_unknown=True):
        """Assembles the relative url for that route and the subdomain.
        If building doesn't work for some reasons `None` is returned.

        :internal:
        """
        tmp = []
        add = tmp.append
        processed = set(self.arguments)
        for is_dynamic, data in self._trace:
            if is_dynamic:
                try:
                    add(self._converters[data].to_url(values[data]))
                except ValidationError:
                    return
                processed.add(data)
            else:
                add(urlquote(data, encoding=self.router.charset, safe='/:|+'))
        domain_part, url = (u''.join(tmp)).split(u'|', 1)

        if append_unknown:
            query_vars = MultiDict(values)
            for key in processed:
                if key in query_vars:
                    del query_vars[key]

            if query_vars:
                url += u'?' + urlencode(
                    sorted(query_vars.items(), key=self.router.sort_key),
                    encoding=self.router.charset,
                )

        return domain_part, url

    def provides_defaults_for(self, route):
        """Check if this route has defaults for a given route.

        :internal:
        """
        return (
            not self.build_only and self.defaults and
            self.endpoint == route.endpoint and self != route and
            self.arguments == route.arguments
        )

    def suitable_for(self, values):
        """Check if the dict of values has enough data for url generation.

        :internal:
        """
        defaults = self.defaults or ()

        # all arguments required must be either in the defaults dict or
        # the value dictionary otherwise it's not suitable
        for key in self.arguments:
            if key not in defaults and key not in values:
                return False

        # in case defaults are given we ensure that either the value was
        # skipped or the value is the same as the default value.
        if defaults:
            for key, value in defaults.items():
                if key in values and value != values[key]:
                    return False

        return True

    def match_compare_key(self):
        """The match compare key for sorting.

        Current implementation:

        1.  routes without any arguments come first for performance
            reasons only as we expect them to match faster and some
            common ones usually don't have any arguments (index pages etc.)
        2.  The more complex routes come first so the second argument is the
            negative length of the number of weights.
        3.  lastly we order by the actual weights.

        :internal:
        """
        return bool(self.arguments), -len(self._weights), self._weights

    def build_compare_key(self):
        """The build compare key for sorting.

        :internal:
        """
        return (
            self.alias and 1 or 0, -len(self.arguments),
            -len(self.defaults or ())
        )

    def __eq__(self, other):
        return (
            self.__class__ is other.__class__ and
            self._trace == other._trace
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return self.route

    def __repr__(self):
        if self.router is None:
            return u'<%s (unbound)>' % self.__class__.__name__
        tmp = []
        for is_dynamic, data in self._trace:
            if is_dynamic:
                tmp.append(u'<%s>' % data)
            else:
                tmp.append(data)
        return u'<%s %s -> %s>' % (
            self.__class__.__name__,
            repr((u''.join(tmp)).lstrip(u'|')).lstrip(u'u'),
            self.endpoint
        )


class BaseConverter(object):
    """Base class for all converters."""
    regex = '[^/]+'
    weight = 100

    def __init__(self, router):
        self.router = router

    def to_python(self, value):
        return value

    def to_url(self, value):
        if not isinstance(value, (str, bytes)):
            value = str(value)
        return urlquote(value, encoding=self.router.charset, safe='/:')


class UnicodeConverter(BaseConverter):
    """This converter is the default converter and accepts any string but
    only one path segment.  Thus the string can not include a slash.

    This is the default validator.

    Example::

        Route('/pages/<page>'),
        Route('/<string(length=2):lang_code>')

    :param router:
        The :class:`URLMap`.
    :param minlength:
        The minimum length of the string.  Must be greater or equal to 1.
    :param maxlength:
        The maximum length of the string.
    :param length:
        The exact length of the string.
    """

    def __init__(self, router, minlength=1, maxlength=None, length=None):
        BaseConverter.__init__(self, router)
        if length is not None:
            length = '{%d}' % int(length)
        else:
            if maxlength is None:
                maxlength = ''
            else:
                maxlength = int(maxlength)
            length = '{%s,%s}' % (
                int(minlength),
                maxlength
            )
        self.regex = '[^/]' + length


class AnyConverter(BaseConverter):
    """Matches one of the items provided.  Items can either be Python
    identifiers or strings::

        Route('/<any(about, help, imprint, class, "foo,bar"):page_name>')

    :param router:
        The :class:`URLMap`.
    :param items:
        This function accepts the possible items as positional arguments.
    """

    def __init__(self, router, *items):
        BaseConverter.__init__(self, router)
        self.regex = '(?:%s)' % '|'.join([re.escape(x) for x in items])


class PathConverter(BaseConverter):
    """Like the default :class:`UnicodeConverter`, but it also matches
    slashes.  This is useful for wikis and similar applications::

        Route('/<path:wikipage>')
        Route('/<path:wikipage>/edit')

    :param router:
        The :class:`URLMap`.
    """
    regex = '[^/].*?'
    weight = 200


class NumberConverter(BaseConverter):
    """Baseclass for `IntegerConverter` and `FloatConverter`.

    :internal:
    """
    weight = 50

    def __init__(self, router, fixed_digits=0, min=None, max=None):
        BaseConverter.__init__(self, router)
        self.fixed_digits = fixed_digits
        self.min = min
        self.max = max

    def num_convert(self, value):
        raise NotImplementedError(
            "num_convert must be overridden by subclasses of NumberConverter"
        )

    def to_python(self, value):
        if (self.fixed_digits and len(value) != self.fixed_digits):
            raise ValidationError()
        value = self.num_convert(value)
        if (
            (self.min is not None and value < self.min) or
            (self.max is not None and value > self.max)
        ):
            raise ValidationError()
        return value

    def to_url(self, value):
        value = self.num_convert(value)
        if self.fixed_digits:
            value = ('%%0%sd' % self.fixed_digits) % value
        return str(value)


class IntegerConverter(NumberConverter):
    """This converter only accepts integer values::

        Route('/page/<int:page>')

    This converter does not support negative values.

    :param router:
        The :class:`URLMap`.
    :param fixed_digits:
        The number of fixed digits in the URL.  If you set this to ``4`` for
        example, the application will only match if the url looks like
        ``/0001/``.  The default is variable length.
    :param min:
        The minimal value.
    :param max:
        The maximal value.
    """
    regex = r'\d+'
    num_convert = int


class FloatConverter(NumberConverter):
    """This converter only accepts floating point values::

        Route('/probability/<float:probability>')

    This converter does not support negative values.

    :param router:
        The :class:`URLMap`.
    :param min:
        The minimal value.
    :param max:
        The maximal value.
    """
    regex = r'\d+\.\d+'
    num_convert = float

    def __init__(self, router, min=None, max=None):
        NumberConverter.__init__(self, router, 0, min, max)


class UUIDConverter(BaseConverter):
    """This converter only accepts UUID strings::

        Route('/object/<uuid:identifier>')

    :param router:
        The :class:`URLMap`.
    """
    regex = (
        r'[A-Fa-f0-9]{8}-[A-Fa-f0-9]{4}-'
        r'[A-Fa-f0-9]{4}-[A-Fa-f0-9]{4}-[A-Fa-f0-9]{12}'
    )

    def to_python(self, value):
        return uuid.UUID(value)

    def to_url(self, value):
        return str(value)


#: the default converter mapping for the router.
DEFAULT_CONVERTERS = {
    'default': UnicodeConverter,
    'string': UnicodeConverter,
    'any': AnyConverter,
    'path': PathConverter,
    'int': IntegerConverter,
    'float': FloatConverter,
    'uuid': UUIDConverter,
}


class URLMap(object):
    """The router class stores all the URL rules and some configuration
    parameters.  Some of the configuration values are only stored on the
    `URLMap` instance since those affect all routes, others are just defaults
    and can be overridden for each route.  Note that you have to specify all
    arguments besides the `routes` as keyword arguments!

    :param routes:
        Sequence of url routes for this router.
    :param default_subdomain:
        The default subdomain for routes without a subdomain defined.
    :param charset:
        Charset of the url (defaults to ``"utf-8"``).
    :param strict_slashes:
        Take care of trailing slashes.
    :param redirect_defaults:
        This will redirect to the default route if it wasn't visited that way.
        This helps creating unique URLs.
    :param converters:
        A dict of converters that adds additional converters to the list of
        converters. If you redefine one converter this will override the
        original one.
    :param sort_parameters:
        If set to `True` the url parameters are sorted.  See `urlencode` for
        more details.
    :param sort_key:
        The sort key function for `urlencode`.
    :param encoding_errors:
        The error method to use for decoding
    :param host_matching:
        If set to `True` it enables the host matching feature and disables the
        subdomain one.  If enabled the `host` parameter to routes is used
        instead of the `subdomain` one.
    """

    default_converters = ImmutableDict(DEFAULT_CONVERTERS)

    def __init__(
        self, routes=None, default_subdomain='', charset='utf-8',
        strict_slashes=True, redirect_defaults=True,
        converters=None, sort_parameters=False, sort_key=None,
        encoding_errors='replace', host_matching=False,
    ):
        self._routes = []
        self._routes_by_endpoint = {}
        self._remap = True

        self.default_subdomain = default_subdomain
        self.charset = charset
        self.encoding_errors = encoding_errors
        self.strict_slashes = strict_slashes
        self.redirect_defaults = redirect_defaults
        self.host_matching = host_matching

        self.converters = self.default_converters.copy()
        if converters:
            self.converters.update(converters)

        self.sort_parameters = sort_parameters
        self.sort_key = sort_key

        self.add_routes(*routes or ())

    def is_endpoint_expecting(self, endpoint, *arguments):
        """Iterate over all routes and check if the endpoint expects
        the arguments provided.  This is for example useful if you have
        some URLs that expect a language code and others that do not and
        you want to wrap the builder a bit so that the current language
        code is automatically added if not provided but endpoints expect
        it.

        :param endpoint:
            The endpoint to check.
        :param arguments:
            This function accepts one or more arguments as positional
            arguments.  Each one of them is checked.
        """
        self.update()
        arguments = set(arguments)
        for route in self._routes_by_endpoint[endpoint]:
            if arguments.issubset(route.arguments):
                return True
        return False

    def iter_routes(self, endpoint=None):
        """Iterate over all routes or the routes of an endpoint.

        :param endpoint:
            If provided only the routes for that endpoint are returned.
        :return: an iterator
        """
        self.update()
        if endpoint is not None:
            return iter(self._routes_by_endpoint[endpoint])
        return iter(self._routes)

    def add_routes(self, *factories):
        """Add a new route or factory to the router and bind it.  Requires that
        the route is not bound to another router.

        :param routefactory:
            A :class:`Route` or :class:`RouteFactory`
        """
        for factory in factories:
            for route in factory.get_routes(self):
                route.bind(self)
                self._routes.append(route)
                self._routes_by_endpoint.setdefault(
                    route.endpoint, []
                ).append(route)
        self._remap = True

    def bind(
        self, server_name, script_name=None, subdomain=None,
        url_scheme='http', path_info=None, query_args=None,
    ):
        """Return a new :class:`MapAdapter` with the details specified to the
        call.  Note that `script_name` will default to ``'/'`` if not further
        specified or `None`.  The `server_name` at least is a requirement
        because the HTTP RFC requires absolute URLs for redirects and so all
        redirect exceptions raised by Verktyg will contain the full canonical
        URL.

        If no path_info is passed to :meth:`match` it will use the default path
        info passed to bind.  While this doesn't really make sense for
        manual bind calls, it's useful if you bind a router to a WSGI
        environment which already contains the path info.

        `subdomain` will default to the `default_subdomain` for this router if
        no defined. If there is no `default_subdomain` you cannot use the
        subdomain feature.
        """
        if self.host_matching:
            if subdomain is not None:
                raise RuntimeError(
                    'host matching enabled and a subdomain was provided'
                )
        elif subdomain is None:
            subdomain = self.default_subdomain

        if subdomain is not None:
            subdomain = subdomain.lower()
            subdomain = subdomain.encode('idna').decode('ascii')

        if server_name is not None:
            server_name = server_name.lower()
            server_name = server_name.encode('idna').decode('ascii')

        if subdomain and not server_name:
            raise TypeError('subdomain specified without server name')

        if script_name is None:
            script_name = '/'

        return MapAdapter(
            self, server_name, script_name, subdomain,
            url_scheme, path_info, query_args,
        )

    def bind_to_environ(self, environ, server_name=None, subdomain=None):
        """Like :meth:`bind` but you can pass it an WSGI environment and it
        will fetch the information from that dictionary.  Note that because of
        limitations in the protocol there is no way to get the current
        subdomain and real `server_name` from the environment.  If you don't
        provide it, Verktyg will use `SERVER_NAME` and `SERVER_PORT` (or
        `HTTP_HOST` if provided) as used `server_name` with disabled subdomain
        feature.

        If `subdomain` is `None` but an environment and a server name is
        provided it will calculate the current subdomain automatically.
        Example: `server_name` is ``'example.com'`` and the `SERVER_NAME`
        in the wsgi `environ` is ``'staging.dev.example.com'`` the calculated
        subdomain will be ``'staging.dev'``.

        If the object passed as environ has an environ attribute, the value of
        this attribute is used instead.  This allows you to pass request
        objects.  Additionally `PATH_INFO` added as a default of the
        :class:`MapAdapter` so that you don't have to pass the path info to
        the match method.

        :param environ:
            A WSGI environment.
        :param server_name:
            An optional server name hint (see above).
        :param subdomain:
            Optionally the current subdomain (see above).
        """
        environ = getattr(environ, 'environ', environ)
        assert isinstance(environ, dict)
        if server_name is None:
            if 'HTTP_HOST' in environ:
                server_name = environ['HTTP_HOST']
            else:
                server_name = environ['SERVER_NAME']
                if (
                    (environ['wsgi.url_scheme'], environ['SERVER_PORT']) not
                    in (('https', '443'), ('http', '80'))
                ):
                    server_name += ':' + environ['SERVER_PORT']
        elif subdomain is None and not self.host_matching:
            server_name = server_name.lower()
            if 'HTTP_HOST' in environ:
                wsgi_server_name = environ.get('HTTP_HOST')
            else:
                wsgi_server_name = environ.get('SERVER_NAME')
                if (
                    (environ['wsgi.url_scheme'], environ['SERVER_PORT']) not
                    in (('https', '443'), ('http', '80'))
                ):
                    wsgi_server_name += ':' + environ['SERVER_PORT']
            wsgi_server_name = wsgi_server_name.lower()
            cur_server_name = wsgi_server_name.split('.')
            real_server_name = server_name.split('.')
            offset = -len(real_server_name)
            if cur_server_name[offset:] != real_server_name:
                # This can happen even with valid configs if the server was
                # accesssed directly by IP address under some situations.
                # Instead of raising an exception like in Werkzeug 0.7 or
                # earlier we go by an invalid subdomain which will result
                # in a 404 error on matching.
                subdomain = '<invalid>'
            else:
                subdomain = '.'.join(filter(None, cur_server_name[:offset]))

        def _get_wsgi_string(name):
            val = environ.get(name)
            if val is not None:
                return wsgi_decoding_dance(val, self.charset)

        script_name = _get_wsgi_string('SCRIPT_NAME')
        path_info = _get_wsgi_string('PATH_INFO')
        query_args = _get_wsgi_string('QUERY_STRING')
        return URLMap.bind(
            self, server_name, script_name,
            subdomain, environ['wsgi.url_scheme'],
            path_info, query_args=query_args,
        )

    def update(self):
        """Called before matching and building to keep the compiled routes
        in the correct order after things changed.
        """
        if self._remap:
            self._routes.sort(key=lambda x: x.match_compare_key())
            for routes in self._routes_by_endpoint.values():
                routes.sort(key=lambda x: x.build_compare_key())
            self._remap = False

    def __repr__(self):
        routes = self.iter_routes()
        return '%s(%s)' % (self.__class__.__name__, pformat(list(routes)))


class MapAdapter(object):
    """Returned by :meth:`URLMap.bind` or :meth:`Router.bind_to_environ` and
    does the URL matching and building based on runtime information.
    """

    def __init__(
        self, router, server_name, script_name, subdomain,
        url_scheme, path_info, query_args=None,
    ):
        self.router = router
        self.server_name = server_name
        script_name = script_name
        if not script_name.endswith(u'/'):
            script_name += u'/'
        self.script_name = script_name
        self.subdomain = subdomain
        self.url_scheme = url_scheme
        self.path_info = path_info
        self.query_args = query_args

    def match(self, path_info=None, return_route=False, query_args=None):
        """The usage is simple: you just pass the match method the current
        path info.  The following things can then happen:

        - you receive a `NotFound` exception that indicates that no URL is
          matching.  A `NotFound` exception is also a WSGI application you
          can call to get a default page not found page (happens to be the
          same object as `verktyg.exceptions.NotFound`)

        - you receive a `RequestRedirect` exception with a `new_url`
          attribute.  This exception is used to notify you about a request
          Werkzeug requests from your WSGI application.  This is for example
          the case if you request ``/foo`` although the correct URL is
          ``/foo/``
          You can use the `RequestRedirect` instance as response-like object
          similar to all other subclasses of `HTTPException`.

        - you get a tuple in the form ``(endpoint, arguments)`` if there is
          a match (unless `return_route` is True, in which case you get a tuple
          in the form ``(route, arguments)``)

        If the path info is not passed to the match method the default path
        info of the map is used (defaults to the root URL if not defined
        explicitly).

        All of the exceptions raised are subclasses of `HTTPException` so they
        can be used as WSGI responses.  The will all render generic error or
        redirect pages.

        Here is a small example for matching:

        >>> m = URLMap([
        ...     Route('/', endpoint='index'),
        ...     Route('/downloads/', endpoint='downloads/index'),
        ...     Route('/downloads/<int:id>', endpoint='downloads/show')
        ... ])
        >>> urls = m.bind("example.com", "/")
        >>> urls.match("/", "GET")
        ('index', {})
        >>> urls.match("/downloads/42")
        ('downloads/show', {'id': 42})

        And here is what happens on redirect and missing URLs:

        >>> urls.match("/downloads")
        Traceback (most recent call last):
          ...
        RequestRedirect: http://example.com/downloads/
        >>> urls.match("/missing")
        Traceback (most recent call last):
          ...
        NotFound: 404 Not Found

        :param path_info:
            The path info to use for matching.  Overrides the path info
            specified on binding.
        :param return_route:
            Return the route that matched instead of just the endpoint
            (defaults to `False`).
        :param query_args:
            Optional query arguments that are used for automatic redirects as
            string or dictionary.  It's currently not possible to use the query
            arguments for URL matching.
        """
        self.router.update()
        if path_info is None:
            path_info = self.path_info
        if query_args is None:
            query_args = self.query_args

        path = u'%s|/%s' % (
            self.router.host_matching and self.server_name or
            self.subdomain, path_info.lstrip('/')
        )

        for route in self.router._routes:
            try:
                rv = route.match(path)
            except RequestSlash:
                raise RequestRedirect(self.make_redirect_url(
                    urlquote(
                        path_info, encoding=self.router.charset, safe='/:|+'
                    ) + '/', query_args
                ))
            except RequestAliasRedirect as e:
                raise RequestRedirect(self.make_alias_redirect_url(
                    path, route.endpoint, e.matched_values, query_args
                ))
            if rv is None:
                continue

            if self.router.redirect_defaults:
                redirect_url = self.get_default_redirect(route, rv, query_args)
                if redirect_url is not None:
                    raise RequestRedirect(redirect_url)

            if route.redirect_to is not None:
                if isinstance(route.redirect_to, str):
                    def _handle_match(match):
                        value = rv[match.group(1)]
                        return route._converters[match.group(1)].to_url(value)
                    redirect_url = _simple_route_re.sub(
                        _handle_match, route.redirect_to,
                    )
                else:
                    redirect_url = route.redirect_to(self, **rv)
                raise RequestRedirect(str(urljoin('%s://%s%s%s' % (
                    self.url_scheme,
                    self.subdomain and self.subdomain + '.' or '',
                    self.server_name,
                    self.script_name
                ), redirect_url)))

            if return_route:
                return route, rv
            else:
                return route.endpoint, rv

        raise NotFound()

    def test(self, path_info=None):
        """Test if a route would match.  Works like `match` but returns `True`
        if the URL matches, or `False` if it does not exist.

        :param path_info:
            The path info to use for matching.  Overrides the path info
            specified on binding.
        """
        try:
            self.match(path_info)
        except RequestRedirect:
            pass
        except HTTPException:
            return False
        return True

    def get_host(self, domain_part):
        """Figures out the full host name for the given domain part.  The
        domain part is a subdomain in case host matching is disabled or
        a full host name.
        """
        if self.router.host_matching:
            if domain_part is None:
                return self.server_name
            return domain_part
        subdomain = domain_part
        if subdomain is None:
            subdomain = self.subdomain

        if subdomain:
            return subdomain + '.' + self.server_name

        return self.server_name

    def get_default_redirect(self, route, values, query_args):
        """A helper that returns the URL to redirect to if it finds one.
        This is used for default redirecting only.

        :internal:
        """
        assert self.router.redirect_defaults
        for r in self.router._routes_by_endpoint[route.endpoint]:
            # every route that comes after this one, including ourself
            # has a lower priority for the defaults.  We order the ones
            # with the highest priority up for building.
            if r is route:
                break
            if (
                r.provides_defaults_for(route) and
                r.suitable_for(values)
            ):
                values.update(r.defaults)
                domain_part, path = r.build(values)
                return self.make_redirect_url(
                    path, query_args, domain_part=domain_part)

    def encode_query_args(self, query_args):
        if not isinstance(query_args, str):
            query_args = urlencode(query_args, self.router.charset)
        return query_args

    def make_redirect_url(self, path_info, query_args=None, domain_part=None):
        """Creates a redirect URL.

        :internal:
        """
        return urlunparse(ParseResult(
            scheme=self.url_scheme or '',
            netloc=self.get_host(domain_part),
            path=posixpath.join(
                self.script_name.strip('/'), path_info.lstrip('/')
            ),
            query=self.encode_query_args(query_args) if query_args else '',
            params=None, fragment=None
        ))

    def make_alias_redirect_url(self, path, endpoint, values, query_args):
        """Internally called to make an alias redirect URL."""
        url = self.build(
            endpoint, values, append_unknown=False, force_external=True,
        )
        if query_args:
            url += '?' + self.encode_query_args(query_args)
        assert url != path, (
            'detected invalid alias setting.  No canonical URL found'
        )
        return url

    def build(
        self, endpoint, values=None, force_external=False,
        append_unknown=True,
    ):
        """Building URLs works pretty much the other way round.  Instead of
        `match` you call `build` and pass it the endpoint and a dict of
        arguments for the placeholders.

        The `build` function also accepts an argument called `force_external`
        which, if you set it to `True` will force external URLs. Per default
        external URLs (include the server name) will only be used if the
        target URL is on a different subdomain.

        >>> m = URLMap([
        ...     Route('/', endpoint='index'),
        ...     Route('/downloads/', endpoint='downloads/index'),
        ...     Route('/downloads/<int:id>', endpoint='downloads/show')
        ... ])
        >>> urls = m.bind("example.com", "/")
        >>> urls.build("index", {})
        '/'
        >>> urls.build("downloads/show", {'id': 42})
        '/downloads/42'
        >>> urls.build("downloads/show", {'id': 42}, force_external=True)
        'http://example.com/downloads/42'

        Because URLs cannot contain non ASCII data you will always get
        bytestrings back.  Non ASCII characters are urlencoded with the
        charset defined on the router instance.

        Additional values are converted to unicode and appended to the URL as
        URL querystring parameters:

        >>> urls.build("index", {'q': 'My Searchstring'})
        '/?q=My+Searchstring'

        If a route does not exist when building a `BuildError` exception is
        raised.

        :param endpoint:
            The endpoint of the URL to build.
        :param values:
            The values for the URL to build.  Unhandled values are appended to
            the URL as query parameters.
        :param force_external:
            Enforce full canonical external URLs.
        :param append_unknown:
            Unknown parameters are appended to the generated URL as query
            string argument.  Disable this if you want the builder to ignore
            those.
        """
        self.router.update()
        if values:
            if isinstance(values, MultiDict):
                valueiter = values.iteritems(multi=True)
            else:
                valueiter = values.items()
            values = dict((k, v) for k, v in valueiter if v is not None)
        else:
            values = {}

        rv = None
        for route in self.router._routes_by_endpoint.get(endpoint, ()):
            if route.suitable_for(values):
                rv = route.build(values, append_unknown)
                if rv is not None:
                    break
        if rv is None:
            raise BuildError(endpoint, values)
        domain_part, path = rv

        host = self.get_host(domain_part)

        # shortcut this.
        if not force_external:
            host_matching = self.router.host_matching
            if (
                (host_matching and host == self.server_name) or
                (not host_matching and domain_part == self.subdomain)
            ):
                return str(urljoin(self.script_name, './' + path.lstrip('/')))

        return urlunparse(ParseResult(
            scheme=self.url_scheme or '',
            netloc=host,
            path=posixpath.join(self.script_name, path.lstrip('/')),
            params=None, query=None, fragment=None
        ))



"""
    verktyg.http.basic
    ~~~~~~~~~~~~~~~~~~

    :copyright:
        (c) 2015 Ben Mather, based on Werkzeug, see AUTHORS for more details.
    :license:
        BSD, see LICENSE for more details.
"""
import re
import sys
from io import BytesIO
from time import gmtime
from email.utils import parsedate_tz
from urllib.request import parse_http_list as _parse_list_header
from datetime import datetime, timedelta

from verktyg.datastructures import is_immutable
from verktyg import exceptions


_missing = object()

_cookie_charset = 'latin1'
# for explanation of "media-range", etc. see Sections 5.3.{1,2} of RFC 7231
_token_chars = frozenset(
    "!#$%&'*+-.0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    '^_`abcdefghijklmnopqrstuvwxyz|~'
)
_unsafe_header_chars = set('()<>@,;:\"/[]?={} \t')
_quoted_string_re = r'"[^"\\]*(?:\\.[^"\\]*)*"'
_option_header_piece_re = re.compile(
    r';\s*(%s|[^\s;=]+)\s*(?:=\s*(%s|[^;]+))?\s*' %
    (_quoted_string_re, _quoted_string_re)
)

_entity_headers = frozenset([
    'allow', 'content-encoding', 'content-language', 'content-length',
    'content-location', 'content-md5', 'content-range', 'content-type',
    'expires', 'last-modified'
])
_hop_by_hop_headers = frozenset([
    'connection', 'keep-alive', 'proxy-authenticate',
    'proxy-authorization', 'te', 'trailer', 'transfer-encoding',
    'upgrade'
])


def wsgi_to_bytes(data):
    """coerce wsgi unicode represented bytes to real ones

    """
    if isinstance(data, bytes):
        return data
    return data.encode('latin1')  # XXX: utf8 fallback?


def bytes_to_wsgi(data):
    assert isinstance(data, bytes), 'data must be bytes'
    if isinstance(data, str):
        return data
    else:
        return data.decode('latin1')


def quote_header_value(value, extra_chars='', allow_token=True):
    """Quote a header value if necessary.

    :param value:
        The value to quote.
    :param extra_chars:
        A list of extra characters to skip quoting.
    :param allow_token:
        If this is enabled token values are returned unchanged.
    """
    if isinstance(value, bytes):
        value = bytes_to_wsgi(value)
    value = str(value)
    if allow_token:
        token_chars = _token_chars | set(extra_chars)
        if set(value).issubset(token_chars):
            return value
    return '"%s"' % value.replace('\\', '\\\\').replace('"', '\\"')


def unquote_header_value(value, is_filename=False):
    r"""Unquotes a header value.  (Reversal of :func:`quote_header_value`).
    This does not use the real unquoting but what browsers are actually
    using for quoting.

    :param value:
        The header value to unquote.
    """
    if value and value[0] == value[-1] == '"':
        # this is not the real unquoting, but fixing this so that the
        # RFC is met will result in bugs with internet explorer and
        # probably some other browsers as well.  IE for example is
        # uploading files with "C:\foo\bar.txt" as filename
        value = value[1:-1]

        # if this is a filename and the starting characters look like
        # a UNC path, then just return the value without quotes.  Using the
        # replace sequence below on a UNC path has the effect of turning
        # the leading double slash into a single slash and then
        # _fix_ie_filename() doesn't work correctly.  See #458.
        if not is_filename or value[:2] != '\\\\':
            return value.replace('\\\\', '\\').replace('\\"', '"')
    return value


def _options_header_vkw(value, kw):
    return dump_options_header(
        value, dict((k.replace('_', '-'), v) for k, v in kw.items())
    )


def unicodify_header_value(value):
    if isinstance(value, bytes):
        value = value.decode('latin-1')
    if not isinstance(value, str):
        value = str(value)
    return value


class Headers(object):
    """An object that stores some headers.  It has a dict-like interface
    but is ordered and can store the same keys multiple times.

    This data structure is useful if you want a nicer way to handle WSGI
    headers which are stored as tuples in a list.

    From Werkzeug 0.3 onwards, the :exc:`KeyError` raised by this class is
    also a subclass of the :class:`~exceptions.BadRequest` HTTP exception
    and will render a page for a ``400 BAD REQUEST`` if caught in a
    catch-all for HTTP exceptions.

    Headers is mostly compatible with the Python
    :class:`wsgiref.headers.Headers` class, with the exception of
    `__getitem__`.  :mod:`wsgiref` will return `None` for
    ``headers['missing']``, whereas :class:`Headers` will raise a
    :class:`KeyError`.

    To create a new :class:`Headers` object pass it a list or dict of headers
    which are used as default values.  This does not reuse the list passed
    to the constructor for internal usage.

    :param defaults:
        The list of default values for the :class:`Headers`.
    """

    def __init__(self, defaults=None):
        self._list = []
        if defaults is not None:
            if isinstance(defaults, (list, Headers)):
                self._list.extend(defaults)
            else:
                self.extend(defaults)

    def __getitem__(self, key, _get_mode=False):
        if not _get_mode:
            if isinstance(key, int):
                return self._list[key]
            elif isinstance(key, slice):
                return self.__class__(self._list[key])
        if not isinstance(key, str):
            raise exceptions.BadRequestKeyError(key)
        ikey = key.lower()
        for k, v in self._list:
            if k.lower() == ikey:
                return v
        # micro optimization: if we are in get mode we will catch that
        # exception one stack level down so we can raise a standard
        # key error instead of our special one.
        if _get_mode:
            raise KeyError()
        raise exceptions.BadRequestKeyError(key)

    def __eq__(self, other):
        return (
            other.__class__ is self.__class__ and
            set(other._list) == set(self._list)
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def get(self, key, default=None, type=None, as_bytes=False):
        """Return the default value if the requested data doesn't exist.
        If `type` is provided and is a callable it should convert the value,
        return it or raise a :exc:`ValueError` if that is not possible.  In
        this case the function will return the default as if the value was not
        found:

        >>> d = Headers([('Content-Length', '42')])
        >>> d.get('Content-Length', type=int)
        42

        If a headers object is bound you must not add unicode strings
        because no encoding takes place.

        :param key:
            The key to be looked up.
        :param default:
            The default value to be returned if the key can't be looked up.  If
            not further specified `None` is returned.
        :param type:
            A callable that is used to cast the value in the :class:`Headers`.
            If a :exc:`ValueError` is raised by this callable the default value
            is returned.
        :param as_bytes:
            Return bytes instead of unicode strings.
        """
        try:
            rv = self.__getitem__(key, _get_mode=True)
        except KeyError:
            return default
        if as_bytes:
            rv = rv.encode('latin1')
        if type is None:
            return rv
        try:
            return type(rv)
        except ValueError:
            return default

    def getlist(self, key, type=None, as_bytes=False):
        """Return the list of items for a given key. If that key is not in the
        :class:`Headers`, the return value will be an empty list.  Just as
        :meth:`get` :meth:`getlist` accepts a `type` parameter.  All items will
        be converted with the callable defined there.

        :param key:
            The key to be looked up.
        :param type:
            A callable that is used to cast the value in the :class:`Headers`.
            If a :exc:`ValueError` is raised by this callable the value will be
            removed from the list.
        :param as_bytes:
            Return bytes instead of unicode strings.
        :return:
            A :class:`list` of all the values for the key.
        """
        ikey = key.lower()
        result = []
        for k, v in self:
            if k.lower() == ikey:
                if as_bytes:
                    v = v.encode('latin1')
                if type is not None:
                    try:
                        v = type(v)
                    except ValueError:
                        continue
                result.append(v)
        return result

    def get_all(self, name):
        """Return a list of all the values for the named field.

        This method is compatible with the :mod:`wsgiref`
        :meth:`~wsgiref.headers.Headers.get_all` method.
        """
        return self.getlist(name)

    def items(self, lower=False):
        for key, value in self:
            if lower:
                key = key.lower()
            yield key, value

    def keys(self, lower=False):
        for key, _ in self.items(lower):
            yield key

    def values(self):
        for _, value in self.items():
            yield value

    def extend(self, iterable):
        """Extend the headers with a dict or an iterable yielding keys and
        values.
        """
        if isinstance(iterable, dict):
            for key, value in iterable.items():
                if isinstance(value, (tuple, list)):
                    for v in value:
                        self.add(key, v)
                else:
                    self.add(key, value)
        else:
            for key, value in iterable:
                self.add(key, value)

    def __delitem__(self, key, _index_operation=True):
        if _index_operation and isinstance(key, (int, slice)):
            del self._list[key]
            return
        key = key.lower()
        new = []
        for k, v in self._list:
            if k.lower() != key:
                new.append((k, v))
        self._list[:] = new

    def remove(self, key):
        """Remove a key.

        :param key:
            The key to be removed.
        """
        return self.__delitem__(key, _index_operation=False)

    def pop(self, key=None, default=_missing):
        """Removes and returns a key or index.

        :param key:
            The key to be popped.  If this is an integer the item at that
            position is removed, if it's a string the value for that key is.
            If the key is omitted or `None` the last item is removed.
        :return:
            An item.
        """
        if key is None:
            return self._list.pop()
        if isinstance(key, int):
            return self._list.pop(key)
        try:
            rv = self[key]
            self.remove(key)
        except KeyError:
            if default is not _missing:
                return default
            raise
        return rv

    def popitem(self):
        """Removes a key or index and returns a (key, value) item."""
        return self.pop()

    def __contains__(self, key):
        """Check if a key is present."""
        try:
            self.__getitem__(key, _get_mode=True)
        except KeyError:
            return False
        return True

    has_key = __contains__

    def __iter__(self):
        """Yield ``(key, value)`` tuples."""
        return iter(self._list)

    def __len__(self):
        return len(self._list)

    def add(self, _key, _value, **kw):
        """Add a new header tuple to the list.

        Keyword arguments can specify additional parameters for the header
        value, with underscores converted to dashes::

        >>> d = Headers()
        >>> d.add('Content-Type', 'text/plain')
        >>> d.add('Content-Disposition', 'attachment', filename='foo.png')

        The keyword argument dumping uses :func:`dump_options_header`
        behind the scenes.
        """
        if kw:
            _value = _options_header_vkw(_value, kw)
        _value = unicodify_header_value(_value)
        self._validate_value(_value)
        self._list.append((_key, _value))

    def _validate_value(self, value):
        if not isinstance(value, str):
            raise TypeError('Value should be unicode.')
        if u'\n' in value or u'\r' in value:
            raise ValueError(
                'Detected newline in header value.  '
                'This is a potential security problem'
            )

    def add_header(self, _key, _value, **_kw):
        """Add a new header tuple to the list.

        An alias for :meth:`add` for compatibility with the :mod:`wsgiref`
        :meth:`~wsgiref.headers.Headers.add_header` method.
        """
        self.add(_key, _value, **_kw)

    def clear(self):
        """Clears all headers."""
        del self._list[:]

    def set(self, _key, _value, **kw):
        """Remove all header tuples for `key` and add a new one.  The newly
        added key either appears at the end of the list if there was no
        entry or replaces the first one.

        Keyword arguments can specify additional parameters for the header
        value, with underscores converted to dashes.  See :meth:`add` for
        more information.

        :param key:
            The key to be inserted.
        :param value:
            The value to be inserted.
        """
        if kw:
            _value = _options_header_vkw(_value, kw)
        _value = unicodify_header_value(_value)
        self._validate_value(_value)
        if not self._list:
            self._list.append((_key, _value))
            return
        listiter = iter(self._list)
        ikey = _key.lower()
        for idx, (old_key, old_value) in enumerate(listiter):
            if old_key.lower() == ikey:
                # replace first ocurrence
                self._list[idx] = (_key, _value)
                break
        else:
            self._list.append((_key, _value))
            return
        self._list[idx + 1:] = [t for t in listiter if t[0].lower() != ikey]

    def setdefault(self, key, value):
        """Returns the value for the key if it is in the dict, otherwise it
        returns `default` and sets that value for `key`.

        :param key:
            The key to be looked up.
        :param default:
            The default value to be returned if the key is not in the dict.  If
            not further specified it's `None`.
        """
        if key in self:
            return self[key]
        self.set(key, value)
        return value

    def __setitem__(self, key, value):
        """Like :meth:`set` but also supports index/slice based setting."""
        if isinstance(key, (slice, int)):
            if isinstance(key, int):
                value = [value]
            value = [(k, unicodify_header_value(v)) for (k, v) in value]
            [self._validate_value(v) for (k, v) in value]
            if isinstance(key, int):
                self._list[key] = value[0]
            else:
                self._list[key] = value
        else:
            self.set(key, value)

    def to_list(self, charset='iso-8859-1'):
        """Convert the headers into a list suitable for WSGI."""
        from warnings import warn
        warn(
            DeprecationWarning('Method removed, use to_wsgi_list instead'),
            stacklevel=2,
        )
        return self.to_wsgi_list()

    def to_wsgi_list(self):
        """Convert the headers into a list suitable for WSGI.

        The values are byte strings in Python 2 converted to latin1 and unicode
        strings in Python 3 for the WSGI server to encode.

        :return: list
        """
        return list(self)

    def copy(self):
        return self.__class__(self._list)

    def __copy__(self):
        return self.copy()

    def __str__(self):
        """Returns formatted headers suitable for HTTP transmission."""
        strs = []
        for key, value in self.to_wsgi_list():
            strs.append('%s: %s' % (key, value))
        strs.append('\r\n')
        return '\r\n'.join(strs)

    def __repr__(self):
        return '%s(%r)' % (
            self.__class__.__name__,
            list(self)
        )


class ImmutableHeadersMixin(object):
    """Makes a :class:`Headers` immutable.  We do not mark them as
    hashable though since the only usecase for this datastructure
    in Werkzeug is a view on a mutable structure.

    :private:
    """

    def __delitem__(self, key):
        is_immutable(self)

    def __setitem__(self, key, value):
        is_immutable(self)
    set = __setitem__

    def add(self, item):
        is_immutable(self)
    remove = add_header = add

    def extend(self, iterable):
        is_immutable(self)

    def insert(self, pos, value):
        is_immutable(self)

    def pop(self, index=-1):
        is_immutable(self)

    def popitem(self):
        is_immutable(self)

    def setdefault(self, key, default):
        is_immutable(self)


def dump_options_header(header, options):
    """The reverse function to :func:`parse_options_header`.

    :param header:
        The header to dump.
    :param options:
        A dict of options to append.
    """
    segments = []
    if header is not None:
        segments.append(header)
    for key, value in options.items():
        if value is None:
            segments.append(key)
        else:
            segments.append('%s=%s' % (key, quote_header_value(value)))
    return '; '.join(segments)


def dump_header(iterable, allow_token=True):
    """Dump an HTTP header again.  This is the reversal of
    :func:`parse_list_header`, :func:`parse_set_header` and
    :func:`parse_dict_header`.  This also quotes strings that include an
    equals sign unless you pass it as dict of key, value pairs.

    >>> dump_header({'foo': 'bar baz'})
    'foo="bar baz"'
    >>> dump_header(('foo', 'bar baz'))
    'foo, "bar baz"'

    :param iterable:
        The iterable or dict of values to quote.
    :param allow_token:
        If set to `False` tokens as values are disallowed.
        See :func:`quote_header_value` for more details.
    """
    if isinstance(iterable, dict):
        items = []
        for key, value in iterable.items():
            if value is None:
                items.append(key)
            else:
                items.append('%s=%s' % (
                    key,
                    quote_header_value(value, allow_token=allow_token)
                ))
    else:
        items = [
            quote_header_value(x, allow_token=allow_token)
            for x in iterable
        ]
    return ', '.join(items)


def parse_list_header(value):
    """Parse lists as described by RFC 2068 Section 2.

    In particular, parse comma-separated lists where the elements of
    the list may include quoted-strings.  A quoted-string could
    contain a comma.  A non-quoted string could have quotes in the
    middle.  Quotes are removed automatically after parsing.

    It basically works like :func:`parse_set_header` just that items
    may appear multiple times and case sensitivity is preserved.

    The return value is a standard :class:`list`:

    >>> parse_list_header('token, "quoted value"')
    ['token', 'quoted value']

    To create a header from the :class:`list` again, use the
    :func:`dump_header` function.

    :param value:
        A string with a list header.
    :return:
        :class:`list`
    """
    result = []
    for item in _parse_list_header(value):
        if item[:1] == item[-1:] == '"':
            item = unquote_header_value(item[1:-1])
        result.append(item)
    return result


def parse_dict_header(value, cls=dict):
    """Parse lists of key, value pairs as described by RFC 2068 Section 2 and
    convert them into a python dict (or any other mapping object created from
    the type with a dict like interface provided by the `cls` arugment):

    >>> d = parse_dict_header('foo="is a fish", bar="as well"')
    >>> type(d) is dict
    True
    >>> sorted(d.items())
    [('bar', 'as well'), ('foo', 'is a fish')]

    If there is no value for a key it will be `None`:

    >>> parse_dict_header('key_without_value')
    {'key_without_value': None}

    To create a header from the :class:`dict` again, use the
    :func:`dump_header` function.

    :param value:
        A string with a dict header.
    :param cls:
        Callable to use for storage of parsed results.
    :return:
        An instance of `cls`
    """
    result = cls()
    if not isinstance(value, str):
        # XXX: validate
        value = bytes_to_wsgi(value)
    for item in _parse_list_header(value):
        if '=' not in item:
            result[item] = None
            continue
        name, value = item.split('=', 1)
        if value[:1] == value[-1:] == '"':
            value = unquote_header_value(value[1:-1])
        result[name] = value
    return result


def parse_options_header(value):
    """Parse a ``Content-Type`` like header into a tuple with the content
    type and the options:

    >>> parse_options_header('text/html; charset=utf8')
    ('text/html', {'charset': 'utf8'})

    This should not be used to parse ``Cache-Control`` like headers that use
    a slightly different format.  For these headers use the
    :func:`parse_dict_header` function.

    :param value:
        The header to parse.
    :return:
        (str, options)
    """
    def _tokenize(string):
        for match in _option_header_piece_re.finditer(string):
            key, value = match.groups()
            key = unquote_header_value(key)
            if value is not None:
                value = unquote_header_value(value, key == 'filename')
            yield key, value

    if not value:
        return '', {}

    parts = _tokenize(';' + value)
    name = next(parts)[0]
    extra = dict(parts)
    return name, extra


def parse_date(value):
    """Parse one of the following date formats into a datetime object:

    .. sourcecode:: text

        Sun, 06 Nov 1994 08:49:37 GMT  ; RFC 822, updated by RFC 1123
        Sunday, 06-Nov-94 08:49:37 GMT ; RFC 850, obsoleted by RFC 1036
        Sun Nov  6 08:49:37 1994       ; ANSI C's asctime() format

    If parsing fails the return value is `None`.

    :param value:
        A string with a supported date format.
    :return:
        A :class:`datetime.datetime` object.
    """
    if value:
        t = parsedate_tz(value.strip())
        if t is not None:
            try:
                year = t[0]
                # unfortunately that function does not tell us if two digit
                # years were part of the string, or if they were prefixed
                # with two zeroes.  So what we do is to assume that 69-99
                # refer to 1900, and everything below to 2000
                if year >= 0 and year <= 68:
                    year += 2000
                elif year >= 69 and year <= 99:
                    year += 1900
                return (
                    datetime(*((year,) + t[1:7])) -
                    timedelta(seconds=t[-1] or 0)
                )
            except (ValueError, OverflowError):
                return None


def _dump_date(d, delim):
    """Used for `http_date` and `cookie_date`."""
    if d is None:
        d = gmtime()
    elif isinstance(d, datetime):
        d = d.utctimetuple()
    elif isinstance(d, (int, float)):
        d = gmtime(d)
    return '%s, %02d%s%s%s%s %02d:%02d:%02d GMT' % (
        ('Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun')[d.tm_wday],
        d.tm_mday, delim,
        (
            'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep',
            'Oct', 'Nov', 'Dec',
        )[d.tm_mon - 1],
        delim, str(d.tm_year), d.tm_hour, d.tm_min, d.tm_sec
    )


def http_date(timestamp=None):
    """Formats the time to match the RFC1123 date format.

    Accepts a floating point number expressed in seconds since the epoch in, a
    datetime object or a timetuple.  All times in UTC.  The :func:`parse_date`
    function can be used to parse such a date.

    Outputs a string in the format ``Wdy, DD Mon YYYY HH:MM:SS GMT``.

    :param timestamp:
        If provided that date is used, otherwise the current.
    """
    return _dump_date(timestamp, ' ')


def remove_entity_headers(headers, allowed=('expires', 'content-location')):
    """Remove all entity headers from a list or :class:`Headers` object.  This
    operation works in-place.  `Expires` and `Content-Location` headers are
    by default not removed.  The reason for this is :rfc:`2616` section
    10.3.5 which specifies some entity headers that should be sent.

    :param headers:
        A list or :class:`Headers` object.
    :param allowed:
        A list of headers that should still be allowed even though they are
        entity headers.
    """
    allowed = set(x.lower() for x in allowed)
    headers[:] = [
        (key, value) for key, value in headers
        if not is_entity_header(key) or key.lower() in allowed
    ]


def remove_hop_by_hop_headers(headers):
    """Remove all HTTP/1.1 "Hop-by-Hop" headers from a list or
    :class:`Headers` object.  This operation works in-place.

    :param headers:
        A list or :class:`Headers` object.
    """
    headers[:] = [
        (key, value) for key, value in headers
        if not is_hop_by_hop_header(key)
    ]


def is_entity_header(header):
    """Check if a header is an entity header.

    :param header:
        The header to test.
    :return:
        `True` if it's an entity header, `False` otherwise.
    """
    return header.lower() in _entity_headers


def is_hop_by_hop_header(header):
    """Check if a header is an HTTP/1.1 "Hop-by-Hop" header.

    :param header:
        The header to test.
    :return:
        `True` if it's an entity header, `False` otherwise.
    """
    return header.lower() in _hop_by_hop_headers


def is_byte_range_valid(start, stop, length):
    """Checks if a given byte content range is valid for the given length.
    """
    if (start is None) != (stop is None):
        return False
    elif start is None:
        return length is None or length >= 0
    elif length is None:
        return 0 <= start < stop
    elif start >= stop:
        return False
    return 0 <= start < length


class FileStorage(object):
    """The :class:`FileStorage` class is a thin wrapper over incoming files.
    It is used by the request object to represent uploaded files.  All the
    attributes of the wrapper stream are proxied by the file storage so
    it's possible to do ``storage.read()`` instead of the long form
    ``storage.stream.read()``.
    """

    def __init__(
        self, stream=None, filename=None, name=None,
        content_type=None, content_length=None,
        headers=None,
    ):
        self.name = name
        self.stream = stream or BytesIO()

        # if no filename is provided we can attempt to get the filename
        # from the stream object passed.  There we have to be careful to
        # skip things like <fdopen>, <stderr> etc.  Python marks these
        # special filenames with angular brackets.
        if filename is None:
            filename = getattr(stream, 'name', None)

            # We want to make sure the filename is always unicode.  This might
            # not be if the name attribute is bytes due to the file being
            # opened from the bytes API.
            if isinstance(filename, bytes):
                filename = filename.decode(
                    sys.getfilesystemencoding(), 'replace'
                )

            if filename and filename[0] == '<' and filename[-1] == '>':
                filename = None

        self.filename = filename
        if headers is None:
            headers = Headers()
        self.headers = headers
        if content_type is not None:
            headers['Content-Type'] = content_type
        if content_length is not None:
            headers['Content-Length'] = str(content_length)

    def _parse_content_type(self):
        if not hasattr(self, '_parsed_content_type'):
            self._parsed_content_type = parse_options_header(self.content_type)

    @property
    def content_type(self):
        """The content-type sent in the header.  Usually not available"""
        return self.headers.get('content-type')

    @property
    def content_length(self):
        """The content-length sent in the header.  Usually not available"""
        return int(self.headers.get('content-length') or 0)

    @property
    def mimetype(self):
        """Like :attr:`content_type`, but without parameters (eg, without
        charset, type etc.) and always lowercase.  For example if the content
        type is ``text/HTML; charset=utf-8`` the mimetype would be
        ``'text/html'``.
        """
        self._parse_content_type()
        return self._parsed_content_type[0].lower()

    @property
    def mimetype_params(self):
        """The mimetype parameters as dict.  For example if the content
        type is ``text/html; charset=utf-8`` the params would be
        ``{'charset': 'utf-8'}``.
        """
        self._parse_content_type()
        return self._parsed_content_type[1]

    def save(self, dst, buffer_size=16384):
        """Save the file to a destination path or file object.  If the
        destination is a file object you have to close it yourself after the
        call.  The buffer size is the number of bytes held in memory during
        the copy process.  It defaults to 16KB.

        For secure file saving also have a look at :func:`secure_filename`.

        :param dst:
            A filename or open file object the uploaded file is saved to.
        :param buffer_size:
            The size of the buffer.  This works the same as the `length`
            parameter of :func:`shutil.copyfileobj`.
        """
        from shutil import copyfileobj
        close_dst = False
        if isinstance(dst, str):
            dst = open(dst, 'wb')
            close_dst = True
        try:
            copyfileobj(self.stream, dst, buffer_size)
        finally:
            if close_dst:
                dst.close()

    def close(self):
        """Close the underlying file if possible."""
        try:
            self.stream.close()
        except Exception:
            pass

    def __nonzero__(self):
        return bool(self.filename)
    __bool__ = __nonzero__

    def __getattr__(self, name):
        return getattr(self.stream, name)

    def __iter__(self):
        return iter(self.readline, '')

    def __repr__(self):
        return '<%s: %r (%r)>' % (
            self.__class__.__name__,
            self.filename,
            self.content_type
        )


class HeaderSet(object):
    """Similar to the :class:`ETags` class this implements a set-like
    structure. Unlike :class:`ETags` this is case insensitive and used for
    vary, allow, and content-language headers.

    If not constructed using the :func:`parse_set_header` function the
    instantiation works like this:

    >>> hs = HeaderSet(['foo', 'bar', 'baz'])
    >>> hs
    HeaderSet(['foo', 'bar', 'baz'])
    """

    def __init__(self, headers=None, on_update=None):
        self._headers = list(headers or ())
        self._set = set([x.lower() for x in self._headers])
        self.on_update = on_update

    def add(self, header):
        """Add a new header to the set."""
        self.update((header,))

    def remove(self, header):
        """Remove a header from the set.  This raises an :exc:`KeyError` if the
        header is not in the set.

        :param header:
            The header to be removed.
        """
        key = header.lower()
        if key not in self._set:
            raise KeyError(header)
        self._set.remove(key)
        for idx, key in enumerate(self._headers):
            if key.lower() == header:
                del self._headers[idx]
                break
        if self.on_update is not None:
            self.on_update(self)

    def update(self, iterable):
        """Add all the headers from the iterable to the set.

        :param iterable:
            Updates the set with the items from the iterable.
        """
        inserted_any = False
        for header in iterable:
            key = header.lower()
            if key not in self._set:
                self._headers.append(header)
                self._set.add(key)
                inserted_any = True
        if inserted_any and self.on_update is not None:
            self.on_update(self)

    def discard(self, header):
        """Like :meth:`remove` but ignores errors.

        :param header:
            The header to be discarded.
        """
        try:
            return self.remove(header)
        except KeyError:
            pass

    def find(self, header):
        """Return the index of the header in the set or return -1 if not found.

        :param header:
            The header to be looked up.
        """
        header = header.lower()
        for idx, item in enumerate(self._headers):
            if item.lower() == header:
                return idx
        return -1

    def index(self, header):
        """Return the index of the header in the set or raise an
        :exc:`IndexError`.

        :param header:
            The header to be looked up.
        """
        rv = self.find(header)
        if rv < 0:
            raise IndexError(header)
        return rv

    def clear(self):
        """Clear the set."""
        self._set.clear()
        del self._headers[:]
        if self.on_update is not None:
            self.on_update(self)

    def as_set(self, preserve_casing=False):
        """Return the set as real python set type.  When calling this, all
        the items are converted to lowercase and the ordering is lost.

        :param preserve_casing:
            If set to `True` the items in the set returned will have the
            original case like in the :class:`HeaderSet`, otherwise they will
            be lowercase.
        """
        if preserve_casing:
            return set(self._headers)
        return set(self._set)

    def to_header(self):
        """Convert the header set into an HTTP header string."""
        return ', '.join(map(quote_header_value, self._headers))

    def __getitem__(self, idx):
        return self._headers[idx]

    def __delitem__(self, idx):
        rv = self._headers.pop(idx)
        self._set.remove(rv.lower())
        if self.on_update is not None:
            self.on_update(self)

    def __setitem__(self, idx, value):
        old = self._headers[idx]
        self._set.remove(old.lower())
        self._headers[idx] = value
        self._set.add(value.lower())
        if self.on_update is not None:
            self.on_update(self)

    def __contains__(self, header):
        return header.lower() in self._set

    def __len__(self):
        return len(self._set)

    def __iter__(self):
        return iter(self._headers)

    def __nonzero__(self):
        return bool(self._set)

    def __str__(self):
        return self.to_header()

    def __repr__(self):
        return '%s(%r)' % (
            self.__class__.__name__,
            self._headers
        )


def parse_set_header(value, on_update=None):
    """Parse a set-like header and return a
    :class:`~HeaderSet` object:

    >>> hs = parse_set_header('token, "quoted value"')

    The return value is an object that treats the items case-insensitively
    and keeps the order of the items:

    >>> 'TOKEN' in hs
    True
    >>> hs.index('quoted value')
    1
    >>> hs
    HeaderSet(['token', 'quoted value'])

    To create a header from the :class:`HeaderSet` again, use the
    :func:`dump_header` function.

    :param value:
        A set header to be parsed.
    :param on_update:
        An optional callable that is called every time a value on the
        :class:`~HeaderSet` object is changed.
    :return:
        A:class:`~HeaderSet`
    """
    if not value:
        return HeaderSet(None, on_update)
    return HeaderSet(parse_list_header(value), on_update)



"""
    verktyg.testsuite.test_http
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Tests for HTTP parsing utilities.

    :copyright:
        (c) 2015 Ben Mather, based on Werkzeug, see AUTHORS for more details.
    :license:
        BSD, see LICENSE for more details.
"""
import unittest

from datetime import datetime

from verktyg.wsgi import wsgi_encoding_dance
from verktyg.test import create_environ

from verktyg import http


class HeadersTestCase(unittest.TestCase):
    storage_class = http.Headers

    def test_basic_interface(self):
        headers = self.storage_class()
        headers.add('Content-Type', 'text/plain')
        headers.add('X-Foo', 'bar')
        self.assertIn('x-Foo', headers)
        self.assertIn('Content-type', headers)

        headers['Content-Type'] = 'foo/bar'
        self.assertEqual(headers['Content-Type'], 'foo/bar')
        self.assertEqual(len(headers.getlist('Content-Type')), 1)

        # list conversion
        self.assertEqual(headers.to_wsgi_list(), [
            ('Content-Type', 'foo/bar'),
            ('X-Foo', 'bar')
        ])
        self.assertEqual(str(headers), (
            "Content-Type: foo/bar\r\n"
            "X-Foo: bar\r\n"
            "\r\n"
        ))
        self.assertEqual(str(self.storage_class()), "\r\n")

        # extended add
        headers.add('Content-Disposition', 'attachment', filename='foo')
        self.assertEqual(
            headers['Content-Disposition'], 'attachment; filename=foo'
        )

        headers.add('x', 'y', z='"')
        self.assertEqual(headers['x'], r'y; z="\""')

    def test_header_set_duplication_bug(self):
        headers = self.storage_class([
            ('Content-Type', 'text/html'),
            ('Foo', 'bar'),
            ('Blub', 'blah')
        ])
        headers['blub'] = 'hehe'
        headers['blafasel'] = 'humm'
        self.assertEqual(headers, self.storage_class([
            ('Content-Type', 'text/html'),
            ('Foo', 'bar'),
            ('blub', 'hehe'),
            ('blafasel', 'humm')
        ]))

    def test_defaults_and_conversion(self):
        # defaults
        headers = self.storage_class([
            ('Content-Type', 'text/plain'),
            ('X-Foo', u'bar'),
            ('X-Bar', u'1'),
            ('X-Bar', u'2')
        ])
        self.assertEqual(headers.getlist('x-bar'), ['1', '2'])
        self.assertEqual(headers.get('x-Bar'), '1')
        self.assertEqual(headers.get('Content-Type'), 'text/plain')

        self.assertEqual(headers.setdefault('X-Foo', 'nope'), 'bar')
        self.assertEqual(headers.setdefault('X-Bar', 'nope'), '1')
        self.assertEqual(headers.setdefault('X-Baz', 'quux'), 'quux')
        self.assertEqual(headers.setdefault('X-Baz', 'nope'), 'quux')
        headers.pop('X-Baz')

        # type conversion
        self.assertEqual(headers.get('x-bar', type=int), 1)
        self.assertEqual(headers.getlist('x-bar', type=int), [1, 2])

        # list like operations
        self.assertEqual(headers[0], ('Content-Type', 'text/plain'))
        self.assertEqual(headers[:1], self.storage_class(
            [('Content-Type', 'text/plain')]
        ))
        del headers[:2]
        del headers[-1]
        self.assertEqual(headers, self.storage_class([('X-Bar', '1')]))

    def test_copying(self):
        a = self.storage_class([('foo', 'bar')])
        b = a.copy()
        a.add('foo', 'baz')
        self.assertEqual(a.getlist('foo'), ['bar', 'baz'])
        self.assertEqual(b.getlist('foo'), ['bar'])

    def test_popping(self):
        headers = self.storage_class([('a', 1)])
        self.assertEqual(headers.pop('a'), 1)
        self.assertEqual(headers.pop('b', 2), 2)

        self.assertRaises(KeyError, headers.pop, 'c')

    def test_set_arguments(self):
        a = self.storage_class()
        a.set('Content-Disposition', 'useless')
        a.set('Content-Disposition', 'attachment', filename='foo')
        self.assertEqual(a['Content-Disposition'], 'attachment; filename=foo')

    def test_reject_newlines(self):
        h = self.storage_class()

        for variation in 'foo\nbar', 'foo\r\nbar', 'foo\rbar':
            try:
                h['foo'] = variation
            except ValueError:
                pass
            else:
                self.fail()

            try:
                h.add('foo', variation)
            except ValueError:
                pass
            else:
                self.fail()

            try:
                h.add('foo', 'test', option=variation)
            except ValueError:
                pass
            else:
                self.fail()

            try:
                h.set('foo', variation)
            except ValueError:
                pass
            else:
                self.fail()

            try:
                h.set('foo', 'test', option=variation)
            except ValueError:
                pass
            else:
                self.fail()

    def test_slicing(self):
        # there's nothing wrong with these being native strings
        # Headers doesn't care about the data types
        h = self.storage_class()
        h.set('X-Foo-Poo', 'bleh')
        h.set('Content-Type', 'application/whocares')
        h.set('X-Forwarded-For', '192.168.0.123')
        h[:] = [(k, v) for k, v in h if k.startswith(u'X-')]
        self.assertEqual(list(h), [
            ('X-Foo-Poo', 'bleh'),
            ('X-Forwarded-For', '192.168.0.123')
        ])

    def test_bytes_operations(self):
        h = self.storage_class()
        h.set('X-Foo-Poo', 'bleh')
        h.set('X-Whoops', b'\xff')

        self.assertEqual(h.get('x-foo-poo', as_bytes=True), b'bleh')
        self.assertEqual(h.get('x-whoops', as_bytes=True), b'\xff')

    def test_to_wsgi_list(self):
        h = self.storage_class()
        h.set(u'Key', u'Value')
        for key, value in h.to_wsgi_list():
            self.assertEqual(key, u'Key')
            self.assertEqual(value, u'Value')


class HTTPUtilityTestCase(unittest.TestCase):
    def test_list_header(self):
        hl = http.parse_list_header('foo baz, blah')
        self.assertEqual(hl, ['foo baz', 'blah'])

    def test_dict_header(self):
        d = http.parse_dict_header('foo="bar baz", blah=42')
        self.assertEqual(d, {'foo': 'bar baz', 'blah': '42'})

    def test_etags_nonzero(self):
        etags = http.parse_etags('w/"foo"')
        assert bool(etags)
        assert etags.contains_raw('w/"foo"')

    def test_parse_date(self):
        self.assertEqual(
            http.parse_date('Sun, 06 Nov 1994 08:49:37 GMT    '),
            datetime(1994, 11, 6, 8, 49, 37)
        )
        self.assertEqual(
            http.parse_date('Sunday, 06-Nov-94 08:49:37 GMT'),
            datetime(1994, 11, 6, 8, 49, 37)
        )
        self.assertEqual(http.parse_date(
            ' Sun Nov  6 08:49:37 1994'),
            datetime(1994, 11, 6, 8, 49, 37)
        )
        self.assertIs(http.parse_date('foo'), None)

    def test_parse_date_overflows(self):
        self.assertEqual(
            http.parse_date(' Sun 02 Feb 1343 08:49:37 GMT'),
            datetime(1343, 2, 2, 8, 49, 37)
        )
        self.assertEqual(
            http.parse_date('Thu, 01 Jan 1970 00:00:00 GMT'),
            datetime(1970, 1, 1, 0, 0)
        )
        self.assertIs(http.parse_date('Thu, 33 Jan 1970 00:00:00 GMT'), None)

    def test_remove_entity_headers(self):
        now = http.http_date()
        headers1 = [
            ('Date', now),
            ('Content-Type', 'text/html'),
            ('Content-Length', '0'),
        ]
        headers2 = http.Headers(headers1)

        http.remove_entity_headers(headers1)
        self.assertEqual(headers1, [('Date', now)])

        http.remove_entity_headers(headers2)
        self.assertEqual(headers2, http.Headers([(u'Date', now)]))

    def test_remove_hop_by_hop_headers(self):
        headers1 = [
            ('Connection', 'closed'),
            ('Foo', 'bar'),
            ('Keep-Alive', 'wtf'),
        ]
        headers2 = http.Headers(headers1)

        http.remove_hop_by_hop_headers(headers1)
        self.assertEqual(headers1, [('Foo', 'bar')])

        http.remove_hop_by_hop_headers(headers2)
        self.assertEqual(headers2, http.Headers([('Foo', 'bar')]))

    def test_parse_options_header(self):
        self.assertEqual(
            http.parse_options_header(r'something; foo="other\"thing"'),
            ('something', {'foo': 'other"thing'})
        )
        self.assertEqual(
            http.parse_options_header(
                r'something; foo="other\"thing"; meh=42'
            ),
            ('something', {'foo': 'other"thing', 'meh': '42'})
        )
        self.assertEqual(
            http.parse_options_header(
                r'something; foo="other\"thing"; meh=42; bleh'
            ),
            ('something', {'foo': 'other"thing', 'meh': '42', 'bleh': None})
        )
        self.assertEqual(
            http.parse_options_header(
                'something; foo="other;thing"; meh=42; bleh'
            ),
            ('something', {'foo': 'other;thing', 'meh': '42', 'bleh': None})
        )
        self.assertEqual(
            http.parse_options_header(
                'something; foo="otherthing"; meh=; bleh'
            ),
            ('something', {'foo': 'otherthing', 'meh': None, 'bleh': None})
        )

    def test_dump_options_header(self):
        self.assertEqual(
            http.dump_options_header('foo', {'bar': 42}), 'foo; bar=42'
        )
        self.assertIn(
            http.dump_options_header('foo', {'bar': 42, 'fizz': None}),
            ('foo; bar=42; fizz', 'foo; fizz; bar=42')
        )

    def test_dump_header(self):
        self.assertEqual(http.dump_header([1, 2, 3]), '1, 2, 3')
        self.assertEqual(
            http.dump_header([1, 2, 3], allow_token=False), '"1", "2", "3"'
        )
        self.assertEqual(
            http.dump_header({'foo': 'bar'}, allow_token=False), 'foo="bar"'
        )
        self.assertEqual(http.dump_header({'foo': 'bar'}), 'foo=bar')

    def test_is_resource_modified(self):
        env = create_environ()

        # ignore POST
        env['REQUEST_METHOD'] = 'POST'
        self.assertFalse(http.is_resource_modified(env, etag='testing'))
        env['REQUEST_METHOD'] = 'GET'

        # etagify from data
        self.assertRaises(
            TypeError, http.is_resource_modified, env, data='42', etag='23'
        )
        env['HTTP_IF_NONE_MATCH'] = http.generate_etag(b'awesome')
        self.assertFalse(http.is_resource_modified(env, data=b'awesome'))

        env['HTTP_IF_MODIFIED_SINCE'] = http.http_date(
            datetime(2008, 1, 1, 12, 30)
        )
        self.assertFalse(
            http.is_resource_modified(
                env, last_modified=datetime(2008, 1, 1, 12, 00)
            )
        )
        self.assertTrue(
            http.is_resource_modified(
                env, last_modified=datetime(2008, 1, 1, 13, 00)
            )
        )

    def test_date_formatting(self):
        self.assertEqual(
            http.cookie_date(0),
            'Thu, 01-Jan-1970 00:00:00 GMT'
        )
        self.assertEqual(
            http.cookie_date(datetime(1970, 1, 1)),
            'Thu, 01-Jan-1970 00:00:00 GMT'
        )
        self.assertEqual(
            http.http_date(0),
            'Thu, 01 Jan 1970 00:00:00 GMT'
        )
        self.assertEqual(
            http.http_date(datetime(1970, 1, 1)),
            'Thu, 01 Jan 1970 00:00:00 GMT'
        )

    def test_cookies(self):
        self.assertEqual(
            dict(
                http.parse_cookie(
                    'dismiss-top=6; CP=null*; '
                    'PHPSESSID=0a539d42abc001cdc762809248d4beed; '
                    'a=42; b="\\\";"'
                )
            ),
            {
                'CP': u'null*',
                'PHPSESSID': u'0a539d42abc001cdc762809248d4beed',
                'a': u'42',
                'dismiss-top': u'6',
                'b': u'\";'
            }
        )
        rv = http.dump_cookie(
            'foo', 'bar baz blub', 360, httponly=True, sync_expires=False
        )
        self.assertIs(type(rv), str)
        self.assertEqual(
            set(rv.split('; ')),
            {
                'HttpOnly',
                'Max-Age=360',
                'Path=/',
                'foo="bar baz blub"'
            }
        )

        self.assertEqual(
            dict(http.parse_cookie('fo234{=bar; blub=Blah')),
            {'fo234{': u'bar', 'blub': u'Blah'}
        )

    def test_cookie_quoting(self):
        val = http.dump_cookie("foo", "?foo")
        self.assertEqual(val, 'foo="?foo"; Path=/')
        self.assertEqual(dict(http.parse_cookie(val)), {'foo': u'?foo'})

        self.assertEqual(
            dict(http.parse_cookie(r'foo="foo\054bar"')),
            {'foo': u'foo,bar'}
        )

    def test_cookie_domain_resolving(self):
        val = http.dump_cookie('foo', 'bar', domain=u'\N{SNOWMAN}.com')
        self.assertEqual(val, 'foo=bar; Domain=xn--n3h.com; Path=/')

    def test_cookie_unicode_dumping(self):
        val = http.dump_cookie('foo', u'\N{SNOWMAN}')
        h = http.Headers()
        h.add('Set-Cookie', val)
        self.assertEqual(h['Set-Cookie'], 'foo="\\342\\230\\203"; Path=/')

        cookies = http.parse_cookie(h['Set-Cookie'])
        self.assertEqual(cookies['foo'], u'\N{SNOWMAN}')

    def test_cookie_unicode_keys(self):
        # Yes, this is technically against the spec but happens
        val = http.dump_cookie(u'fö', u'fö')
        self.assertEqual(
            val, wsgi_encoding_dance(u'fö="f\\303\\266"; Path=/', 'utf-8')
        )
        cookies = http.parse_cookie(val)
        self.assertEqual(cookies[u'fö'], u'fö')

    def test_cookie_unicode_parsing(self):
        # This is actually a correct test.  This is what is being submitted
        # by firefox if you set an unicode cookie and we get the cookie sent
        # in on Python 3 under PEP 3333.
        cookies = http.parse_cookie(u'fÃ¶=fÃ¶')
        self.assertEqual(cookies[u'fö'], u'fö')

    def test_cookie_domain_encoding(self):
        val = http.dump_cookie('foo', 'bar', domain=u'\N{SNOWMAN}.com')
        self.assertEqual(val, 'foo=bar; Domain=xn--n3h.com; Path=/')

        val = http.dump_cookie('foo', 'bar', domain=u'.\N{SNOWMAN}.com')
        self.assertEqual(val, 'foo=bar; Domain=.xn--n3h.com; Path=/')

        val = http.dump_cookie('foo', 'bar', domain=u'.foo.com')
        self.assertEqual(val, 'foo=bar; Domain=.foo.com; Path=/')


class IfRangeTestCase(unittest.TestCase):
    def test_parse_if_range_header(self):
        rv = http.parse_if_range_header('"Test"')
        self.assertEqual(rv.etag, 'Test')
        self.assertIs(rv.date, None)
        self.assertEqual(rv.to_header(), '"Test"')

        # weak information is dropped
        rv = http.parse_if_range_header('w/"Test"')
        self.assertEqual(rv.etag, 'Test')
        self.assertIs(rv.date, None)
        self.assertEqual(rv.to_header(), '"Test"')

        # broken etags are supported too
        rv = http.parse_if_range_header('bullshit')
        self.assertEqual(rv.etag, 'bullshit')
        self.assertIs(rv.date, None)
        self.assertEqual(rv.to_header(), '"bullshit"')

        rv = http.parse_if_range_header('Thu, 01 Jan 1970 00:00:00 GMT')
        self.assertIs(rv.etag, None)
        self.assertEqual(rv.date, datetime(1970, 1, 1))
        self.assertEqual(rv.to_header(), 'Thu, 01 Jan 1970 00:00:00 GMT')

        for x in '', None:
            rv = http.parse_if_range_header(x)
            self.assertIs(rv.etag, None)
            self.assertIs(rv.date, None)
            self.assertEqual(rv.to_header(), '')


class RangeTestCase(unittest.TestCase):
    def test_parse_range(self):
        rv = http.parse_range_header('bytes=52')
        self.assertIs(rv, None)

        rv = http.parse_range_header('bytes=52-')
        self.assertEqual(rv.units, 'bytes')
        self.assertEqual(rv.ranges, [(52, None)])
        self.assertEqual(rv.to_header(), 'bytes=52-')

        rv = http.parse_range_header('bytes=52-99')
        self.assertEqual(rv.units, 'bytes')
        self.assertEqual(rv.ranges, [(52, 100)])
        self.assertEqual(rv.to_header(), 'bytes=52-99')

        rv = http.parse_range_header('bytes=52-99,-1000')
        self.assertEqual(rv.units, 'bytes')
        self.assertEqual(rv.ranges, [(52, 100), (-1000, None)])
        self.assertEqual(rv.to_header(), 'bytes=52-99,-1000')

        rv = http.parse_range_header('bytes = 1 - 100')
        self.assertEqual(rv.units, 'bytes')
        self.assertEqual(rv.ranges, [(1, 101)])
        self.assertEqual(rv.to_header(), 'bytes=1-100')

        rv = http.parse_range_header('AWesomes=0-999')
        self.assertEqual(rv.units, 'awesomes')
        self.assertEqual(rv.ranges, [(0, 1000)])
        self.assertEqual(rv.to_header(), 'awesomes=0-999')


class ContentRangeTestCase(unittest.TestCase):
    def test_parse_content_range(self):
        rv = http.parse_content_range_header('bytes 0-98/*')
        self.assertEqual(rv.units, 'bytes')
        self.assertEqual(rv.start, 0)
        self.assertEqual(rv.stop, 99)
        self.assertIs(rv.length, None)
        self.assertEqual(rv.to_header(), 'bytes 0-98/*')

        rv = http.parse_content_range_header('bytes 0-98/*asdfsa')
        self.assertIs(rv, None)

        rv = http.parse_content_range_header('bytes 0-99/100')
        self.assertEqual(rv.to_header(), 'bytes 0-99/100')
        rv.start = None
        rv.stop = None
        self.assertEqual(rv.units, 'bytes')
        self.assertEqual(rv.to_header(), 'bytes */100')

        rv = http.parse_content_range_header('bytes */100')
        self.assertIs(rv.start, None)
        self.assertIs(rv.stop, None)
        self.assertEqual(rv.length, 100)
        self.assertEqual(rv.units, 'bytes')


class HeaderSetTestCase(unittest.TestCase):
    def test_basic_interface(self):
        hs = http.HeaderSet()
        hs.add('foo')
        hs.add('bar')
        self.assertIn('Bar', hs)
        self.assertEqual(hs.find('foo'), 0)
        self.assertEqual(hs.find('BAR'), 1)
        self.assertLess(hs.find('baz'), 0)
        hs.discard('missing')
        hs.discard('foo')
        self.assertLess(hs.find('foo'), 0)
        self.assertEqual(hs.find('bar'), 0)

        self.assertRaises(IndexError, hs.index, 'missing')

        self.assertEqual(hs.index('bar'), 0)
        self.assertTrue(hs)
        hs.clear()
        self.assertFalse(hs)

    def test_parse_set_header(self):
        hs = http.parse_set_header('foo, Bar, "Blah baz", Hehe')
        self.assertIn('blah baz', hs)
        self.assertNotIn('foobar', hs)
        self.assertIn('foo', hs)
        self.assertEqual(list(hs), ['foo', 'Bar', 'Blah baz', 'Hehe'])
        hs.add('Foo')
        self.assertEqual(hs.to_header(), 'foo, Bar, "Blah baz", Hehe')


class CacheControlTestCase(unittest.TestCase):
    def test_repr(self):
        cc = http.RequestCacheControl(
            [("max-age", "0"), ("private", "True")],
        )
        self.assertEqual(
            repr(cc), "<RequestCacheControl max-age='0' private='True'>"
        )

    def test_cache_control_header(self):
        cc = http.parse_cache_control_header('max-age=0, no-cache')
        self.assertEqual(cc.max_age, 0)
        self.assertTrue(cc.no_cache)
        cc = http.parse_cache_control_header(
            'private, community="UCI"', None,
            http.ResponseCacheControl
        )
        self.assertTrue(cc.private)
        self.assertEqual(cc['community'], 'UCI')

        c = http.ResponseCacheControl()
        self.assertIs(c.no_cache, None)
        self.assertIs(c.private, None)
        c.no_cache = True
        self.assertEqual(c.no_cache, '*')
        c.private = True
        self.assertEqual(c.private, '*')
        del c.private
        self.assertIs(c.private, None)
        self.assertEqual(c.to_header(), 'no-cache')


class ETagsTestCase(unittest.TestCase):
    def test_parse_etags(self):
        self.assertEqual(http.quote_etag('foo'), '"foo"')
        self.assertEqual(http.quote_etag('foo', True), 'w/"foo"')
        self.assertEqual(http.unquote_etag('"foo"'), ('foo', False))
        self.assertEqual(http.unquote_etag('w/"foo"'), ('foo', True))
        es = http.parse_etags('"foo", "bar", w/"baz", blar')
        self.assertEqual(sorted(es), ['bar', 'blar', 'foo'])
        self.assertIn('foo', es)
        self.assertNotIn('baz', es)
        assert es.contains_weak('baz')
        self.assertIn('blar', es)
        assert es.contains_raw('w/"baz"')
        assert es.contains_raw('"foo"')
        self.assertEqual(
            sorted(es.to_header().split(', ')),
            ['"bar"', '"blar"', '"foo"', 'w/"baz"']
        )


class AuthorizationTestCase(unittest.TestCase):
    def test_parse_authorization_header(self):
        a = http.parse_authorization_header(
            'Basic QWxhZGRpbjpvcGVuIHNlc2FtZQ=='
        )
        self.assertEqual(a.type, 'basic')
        self.assertEqual(a.username, 'Aladdin')
        self.assertEqual(a.password, 'open sesame')

        a = http.parse_authorization_header(
            'Digest username="Mufasa", '
            'realm="testrealm@host.invalid", '
            'nonce="dcd98b7102dd2f0e8b11d0f600bfb0c093", '
            'uri="/dir/index.html", '
            'qop=auth, '
            'nc=00000001, '
            'cnonce="0a4f113b", '
            'response="6629fae49393a05397450978507c4ef1", '
            'opaque="5ccc069c403ebaf9f0171e9517f40e41"'
        )
        self.assertEqual(a.type, 'digest')
        self.assertEqual(a.username, 'Mufasa')
        self.assertEqual(a.realm, 'testrealm@host.invalid')
        self.assertEqual(a.nonce, 'dcd98b7102dd2f0e8b11d0f600bfb0c093')
        self.assertEqual(a.uri, '/dir/index.html')
        self.assertIn('auth', a.qop)
        self.assertEqual(a.nc, '00000001')
        self.assertEqual(a.cnonce, '0a4f113b')
        self.assertEqual(a.response, '6629fae49393a05397450978507c4ef1')
        self.assertEqual(a.opaque, '5ccc069c403ebaf9f0171e9517f40e41')

        a = http.parse_authorization_header(
            'Digest username="Mufasa", '
            'realm="testrealm@host.invalid", '
            'nonce="dcd98b7102dd2f0e8b11d0f600bfb0c093", '
            'uri="/dir/index.html", '
            'response="e257afa1414a3340d93d30955171dd0e", '
            'opaque="5ccc069c403ebaf9f0171e9517f40e41"'
        )
        self.assertEqual(a.type, 'digest')
        self.assertEqual(a.username, 'Mufasa')
        self.assertEqual(a.realm, 'testrealm@host.invalid')
        self.assertEqual(a.nonce, 'dcd98b7102dd2f0e8b11d0f600bfb0c093')
        self.assertEqual(a.uri, '/dir/index.html')
        self.assertEqual(a.response, 'e257afa1414a3340d93d30955171dd0e')
        self.assertEqual(a.opaque, '5ccc069c403ebaf9f0171e9517f40e41')

        self.assertIs(http.parse_authorization_header(''), None)
        self.assertIs(http.parse_authorization_header(None), None)
        self.assertIs(http.parse_authorization_header('foo'), None)


class WWWAuthenticateTestCase(unittest.TestCase):
    def test_parse_www_authenticate_header(self):
        wa = http.parse_www_authenticate_header('Basic realm="WallyWorld"')
        self.assertEqual(wa.type, 'basic')
        self.assertEqual(wa.realm, 'WallyWorld')
        wa.realm = 'Foo Bar'
        self.assertEqual(wa.to_header(), 'Basic realm="Foo Bar"')

        wa = http.parse_www_authenticate_header(
            'Digest '
            'realm="testrealm@host.com", '
            'qop="auth,auth-int", '
            'nonce="dcd98b7102dd2f0e8b11d0f600bfb0c093", '
            'opaque="5ccc069c403ebaf9f0171e9517f40e41"'
        )
        self.assertEqual(wa.type, 'digest')
        self.assertEqual(wa.realm, 'testrealm@host.com')
        self.assertIn('auth', wa.qop)
        self.assertIn('auth-int', wa.qop)
        self.assertEqual(wa.nonce, 'dcd98b7102dd2f0e8b11d0f600bfb0c093')
        self.assertEqual(wa.opaque, '5ccc069c403ebaf9f0171e9517f40e41')

        wa = http.parse_www_authenticate_header('broken')
        self.assertEqual(wa.type, 'broken')

        assert not http.parse_www_authenticate_header('').type
        assert not http.parse_www_authenticate_header('')


class FileStorageTestCase(object):
    def test_mimetype_always_lowercase(self):
        file_storage = http.FileStorage(content_type='APPLICATION/JSON')
        self.assertEqual(file_storage.mimetype, 'application/json')



"""
    tests.wsgi
    ~~~~~~~~~~

    Tests the WSGI utilities.

    :copyright:
        (c) 2014 Ben Mather, based on Werkzeug, see AUTHORS for more details.
    :license:
        BSD, see LICENSE for more details.
"""
import unittest

from os import path
from io import StringIO, BytesIO
from tempfile import TemporaryDirectory
from contextlib import closing

from verktyg.test import create_environ, run_wsgi_app
from verktyg.exceptions import BadRequest, ClientDisconnected
from verktyg import wsgi


class WsgiTestCase(unittest.TestCase):
    def test_shareddatamiddleware_get_file_loader(self):
        app = wsgi.SharedDataMiddleware(None, {})
        self.assertTrue(callable(app.get_file_loader('foo')))

    def test_shared_data_middleware(self):
        def null_application(environ, start_response):
            start_response('404 NOT FOUND', [('Content-Type', 'text/plain')])
            yield b'NOT FOUND'

        with TemporaryDirectory() as test_dir:
            test_file_name = path.join(test_dir, 'äöü').encode()
            with open(test_file_name, 'w') as test_file:
                test_file.write('FOUND')

            app = wsgi.SharedDataMiddleware(null_application, {
                '/': path.join(path.dirname(__file__), 'resources'),
                '/sources': path.join(path.dirname(__file__), 'resources'),
                '/foo': test_dir
            })

            for p in '/test.txt', '/sources/test.txt', '/foo/äöü':
                with self.subTest(path=p):
                    app_iter, status, headers = run_wsgi_app(
                        app, create_environ(p)
                    )
                    self.assertEqual(status, '200 OK')
                    with closing(app_iter) as app_iter:
                        data = b''.join(app_iter).strip()
                    self.assertEqual(data, b'FOUND')

            app_iter, status, headers = run_wsgi_app(
                app, create_environ('/missing'))
            self.assertEqual(status, '404 NOT FOUND')
            self.assertEqual(b''.join(app_iter).strip(), b'NOT FOUND')

    def test_dispatchermiddleware(self):
        def null_application(environ, start_response):
            start_response('404 NOT FOUND', [('Content-Type', 'text/plain')])
            yield b'NOT FOUND'

        def dummy_application(environ, start_response):
            start_response('200 OK', [('Content-Type', 'text/plain')])
            yield environ['SCRIPT_NAME'].encode('ascii')

        app = wsgi.DispatcherMiddleware(null_application, {
            '/test1': dummy_application,
            '/test2/very': dummy_application,
        })
        tests = {
            '/test1': ('/test1', '/test1/asfd', '/test1/very'),
            '/test2/very': (
                '/test2/very', '/test2/very/long/path/after/script/name'
            ),
        }
        for name, urls in tests.items():
            for p in urls:
                environ = create_environ(p)
                app_iter, status, headers = run_wsgi_app(app, environ)
                self.assertEqual(status, '200 OK')
                self.assertEqual(
                    b''.join(app_iter).strip(), name.encode('ascii')
                )

        app_iter, status, headers = run_wsgi_app(
            app, create_environ('/missing'))
        self.assertEqual(status, '404 NOT FOUND')
        self.assertEqual(b''.join(app_iter).strip(), b'NOT FOUND')

    def test_get_host(self):
        env = {
            'HTTP_X_FORWARDED_HOST': 'example.org',
            'SERVER_NAME': 'bullshit', 'HOST_NAME': 'ignore me dammit',
        }
        self.assertEqual(wsgi.get_host(env), 'example.org')
        self.assertEqual(
            wsgi.get_host(create_environ('/', 'http://example.org')),
            'example.org'
        )

    def test_get_host_multiple_forwarded(self):
        env = {
            'HTTP_X_FORWARDED_HOST': 'example.com, example.org',
            'SERVER_NAME': 'bullshit', 'HOST_NAME': 'ignore me dammit',
        }
        self.assertEqual(wsgi.get_host(env), 'example.com')
        self.assertEqual(
            wsgi.get_host(create_environ('/', 'http://example.com')),
            'example.com'
        )

    def test_get_host_validation(self):
        env = {
            'HTTP_X_FORWARDED_HOST': 'example.org',
            'SERVER_NAME': 'bullshit', 'HOST_NAME': 'ignore me dammit',
        }
        self.assertEqual(
            wsgi.get_host(env, trusted_hosts=['.example.org']), 'example.org'
        )
        self.assertRaises(
            BadRequest, wsgi.get_host, env, trusted_hosts=['example.com']
        )

    def test_path_info_and_script_name_fetching(self):
        env = create_environ('/\N{SNOWMAN}', 'http://example.com/\N{COMET}/')
        self.assertEqual(
            wsgi.get_path_info(env),
            '/\N{SNOWMAN}'
        )
        self.assertEqual(
            wsgi.get_path_info(env, charset=None),
            '/\N{SNOWMAN}'.encode('utf-8')
        )
        self.assertEqual(
            wsgi.get_script_name(env),
            '/\N{COMET}'
        )
        self.assertEqual(
            wsgi.get_script_name(env, charset=None),
            '/\N{COMET}'.encode('utf-8')
        )

    def test_query_string_fetching(self):
        env = create_environ('/?\N{SNOWMAN}=\N{COMET}')
        qs = wsgi.get_query_string(env)
        self.assertEqual(qs, '%E2%98%83=%E2%98%84')

    def test_limited_stream(self):
        class RaisingLimitedStream(wsgi.LimitedStream):
            def on_exhausted(self):
                raise BadRequest('input stream exhausted')

        io = BytesIO(b'123456')
        stream = RaisingLimitedStream(io, 3)
        self.assertEqual(stream.read(), b'123')
        self.assertRaises(BadRequest, stream.read)

        io = BytesIO(b'123456')
        stream = RaisingLimitedStream(io, 3)
        self.assertEqual(stream.tell(), 0)
        self.assertEqual(stream.read(1), b'1')
        self.assertEqual(stream.tell(), 1)
        self.assertEqual(stream.read(1), b'2')
        self.assertEqual(stream.tell(), 2)
        self.assertEqual(stream.read(1), b'3')
        self.assertEqual(stream.tell(), 3)
        self.assertRaises(BadRequest, stream.read)

        io = BytesIO(b'123456\nabcdefg')
        stream = wsgi.LimitedStream(io, 9)
        self.assertEqual(stream.readline(), b'123456\n')
        self.assertEqual(stream.readline(), b'ab')

        io = BytesIO(b'123456\nabcdefg')
        stream = wsgi.LimitedStream(io, 9)
        self.assertEqual(stream.readlines(), [b'123456\n', b'ab'])

        io = BytesIO(b'123456\nabcdefg')
        stream = wsgi.LimitedStream(io, 9)
        self.assertEqual(stream.readlines(2), [b'12'])
        self.assertEqual(stream.readlines(2), [b'34'])
        self.assertEqual(stream.readlines(), [b'56\n', b'ab'])

        io = BytesIO(b'123456\nabcdefg')
        stream = wsgi.LimitedStream(io, 9)
        self.assertEqual(stream.readline(100), b'123456\n')

        io = BytesIO(b'123456\nabcdefg')
        stream = wsgi.LimitedStream(io, 9)
        self.assertEqual(stream.readlines(100), [b'123456\n', b'ab'])

        io = BytesIO(b'123456')
        stream = wsgi.LimitedStream(io, 3)
        self.assertEqual(stream.read(1), b'1')
        self.assertEqual(stream.read(1), b'2')
        self.assertEqual(stream.read(), b'3')
        self.assertEqual(stream.read(), b'')

        io = BytesIO(b'123456')
        stream = wsgi.LimitedStream(io, 3)
        self.assertEqual(stream.read(-1), b'123')

        io = BytesIO(b'123456')
        stream = wsgi.LimitedStream(io, 0)
        self.assertEqual(stream.read(-1), b'')

        io = StringIO('123456')
        stream = wsgi.LimitedStream(io, 0)
        self.assertEqual(stream.read(-1), '')

        io = StringIO('123\n456\n')
        stream = wsgi.LimitedStream(io, 8)
        self.assertEqual(list(stream), ['123\n', '456\n'])

    def test_limited_stream_disconnection(self):
        io = BytesIO(b'A bit of content')

        # disconnect detection on out of bytes
        stream = wsgi.LimitedStream(io, 255)
        self.assertRaises(ClientDisconnected, stream.read)

        # disconnect detection because file close
        io = BytesIO(b'x' * 255)
        io.close()
        stream = wsgi.LimitedStream(io, 255)
        self.assertRaises(ClientDisconnected, stream.read)

    def test_get_host_fallback(self):
        self.assertEqual(
            wsgi.get_host({
                'SERVER_NAME': 'foobar.example.com',
                'wsgi.url_scheme': 'http',
                'SERVER_PORT': '80'
            }),
            'foobar.example.com'
        )
        self.assertEqual(
            wsgi.get_host({
                'SERVER_NAME': 'foobar.example.com',
                'wsgi.url_scheme': 'http',
                'SERVER_PORT': '81'
            }),
            'foobar.example.com:81'
        )

    def test_get_current_url_unicode(self):
        env = create_environ()
        env['QUERY_STRING'] = 'foo=bar&baz=blah&meh=\xcf'
        rv = wsgi.get_current_url(env)
        self.assertEqual(
            rv, 'http://localhost/?foo=bar&baz=blah&meh=\ufffd'
        )

    def test_multi_part_line_breaks(self):
        data = 'abcdef\r\nghijkl\r\nmnopqrstuvwxyz\r\nABCDEFGHIJK'
        test_stream = StringIO(data)
        lines = list(wsgi.make_line_iter(
            test_stream, limit=len(data), buffer_size=16,
        ))
        self.assertEqual(
            lines, [
                'abcdef\r\n', 'ghijkl\r\n', 'mnopqrstuvwxyz\r\n', 'ABCDEFGHIJK'
            ]
        )

        data = (
            'abc\r\nThis line is broken by the buffer length.'
            '\r\nFoo bar baz'
        )
        test_stream = StringIO(data)
        lines = list(wsgi.make_line_iter(
            test_stream, limit=len(data), buffer_size=24,
        ))
        self.assertEqual(
            lines, [
                'abc\r\n',
                'This line is broken by the buffer length.\r\n',
                'Foo bar baz',
            ]
        )

    def test_multi_part_line_breaks_problematic(self):
        data = 'abc\rdef\r\nghi'
        for x in range(1, 10):
            test_stream = StringIO(data)
            lines = list(wsgi.make_line_iter(
                test_stream, limit=len(data),
                buffer_size=4
            ))
            self.assertEqual(lines, ['abc\r', 'def\r\n', 'ghi'])

    def test_iter_functions_support_iterators(self):
        data = ['abcdef\r\nghi', 'jkl\r\nmnopqrstuvwxyz\r', '\nABCDEFGHIJK']
        lines = list(wsgi.make_line_iter(data))
        self.assertEqual(
            lines, [
                'abcdef\r\n', 'ghijkl\r\n', 'mnopqrstuvwxyz\r\n', 'ABCDEFGHIJK'
            ]
        )

    def test_make_chunk_iter(self):
        data = ['abcdefXghi', 'jklXmnopqrstuvwxyzX', 'ABCDEFGHIJK']
        rv = list(wsgi.make_chunk_iter(data, 'X'))
        self.assertEqual(
            rv, ['abcdef', 'ghijkl', 'mnopqrstuvwxyz', 'ABCDEFGHIJK'],
        )

        data = 'abcdefXghijklXmnopqrstuvwxyzXABCDEFGHIJK'
        test_stream = StringIO(data)
        rv = list(wsgi.make_chunk_iter(
            test_stream, 'X', limit=len(data),
            buffer_size=4,
        ))
        self.assertEqual(
            rv, ['abcdef', 'ghijkl', 'mnopqrstuvwxyz', 'ABCDEFGHIJK']
        )

    def test_make_chunk_iter_bytes(self):
        data = [b'abcdefXghi', b'jklXmnopqrstuvwxyzX', b'ABCDEFGHIJK']
        rv = list(wsgi.make_chunk_iter(data, b'X'))
        self.assertEqual(
            rv, [b'abcdef', b'ghijkl', b'mnopqrstuvwxyz', b'ABCDEFGHIJK']
        )

        data = b'abcdefXghijklXmnopqrstuvwxyzXABCDEFGHIJK'
        test_stream = BytesIO(data)
        rv = list(wsgi.make_chunk_iter(
            test_stream, b'X', limit=len(data), buffer_size=4)
        )
        self.assertEqual(
            rv, [b'abcdef', b'ghijkl', b'mnopqrstuvwxyz', b'ABCDEFGHIJK']
        )

    def test_lines_longer_buffer_size(self):
        data = '1234567890\n1234567890\n'
        for bufsize in range(1, 15):
            lines = list(wsgi.make_line_iter(
                StringIO(data), limit=len(data), buffer_size=4
            ))
            self.assertEqual(lines, ['1234567890\n', '1234567890\n'])


class EnvironHeadersTestCase(unittest.TestCase):
    def test_basic_interface(self):
        # this happens in multiple WSGI servers because they
        # use a vary naive way to convert the headers;
        broken_env = {
            'HTTP_CONTENT_TYPE': 'text/html',
            'CONTENT_TYPE': 'text/html',
            'HTTP_CONTENT_LENGTH': '0',
            'CONTENT_LENGTH': '0',
            'HTTP_ACCEPT': '*',
            'wsgi.version': (1, 0)
        }
        headers = wsgi.EnvironHeaders(broken_env)
        self.assertTrue(headers)
        self.assertEqual(len(headers), 3)
        self.assertEqual(sorted(headers), [
            ('Accept', '*'),
            ('Content-Length', '0'),
            ('Content-Type', 'text/html')
        ])
        self.assertFalse(wsgi.EnvironHeaders({'wsgi.version': (1, 0)}))
        self.assertEqual(len(wsgi.EnvironHeaders({'wsgi.version': (1, 0)})), 0)

    def test_return_type_is_unicode(self):
        # environ contains native strings; we return unicode
        headers = wsgi.EnvironHeaders({
            'HTTP_FOO': '\xe2\x9c\x93',
            'CONTENT_TYPE': 'text/plain',
        })
        self.assertEqual(headers['Foo'], u"\xe2\x9c\x93")
        self.assertIsInstance(headers['Foo'], str)
        self.assertIsInstance(headers['Content-Type'], str)
        iter_output = dict(iter(headers))
        self.assertEqual(iter_output['Foo'], u"\xe2\x9c\x93")
        self.assertIsInstance(iter_output['Foo'], str)
        self.assertIsInstance(iter_output['Content-Type'], str)

    def test_bytes_operations(self):
        foo_val = '\xff'
        h = wsgi.EnvironHeaders({
            'HTTP_X_FOO': foo_val
        })

        self.assertEqual(h.get('x-foo', as_bytes=True), b'\xff')
        self.assertEqual(h.get('x-foo'), u'\xff')



"""
    verktyg.application
    ~~~~~~~~~~~~~~~~~~~

    :copyright: (c) 2014 by Ben Mather.
    :license: BSD, see LICENSE for more details.
"""
import sys
from urllib.parse import urlparse


from verktyg.datastructures import ImmutableDict
from verktyg.exception_dispatch import (
    ExceptionDispatcher, ExceptionHandler
)
from verktyg.utils import redirect
from verktyg.routing import URLMap, Route, RequestRedirect
from verktyg.dispatch import Dispatcher
from verktyg.views import expose
from verktyg.requests import BaseRequest
from verktyg import requests


def _default_redirect_handler(app, req, exc_type, exc_value, exc_traceback):
    return redirect(exc_value.new_url, exc_value.code)


class BaseApplication(object):
    def __init__(
        self, app_root, config, *,
        routes, converters, bindings, handlers,
        middleware, request_class
    ):
        self.config = ImmutableDict(config)

        self._url_map = URLMap(routes, converters=converters)
        root_components = urlparse(app_root)
        self._map_adapter = self._url_map.bind(
            url_scheme=root_components.scheme,
            server_name=root_components.netloc,
            script_name=root_components.path,
        )

        self._dispatcher = Dispatcher(bindings)
        self._exception_dispatcher = ExceptionDispatcher(handlers)

        self.request_class = request_class

        # reference to the bottom of a stack of wsgi middleware wrapping
        # :method:`_dispatch_request`. Invoked by :method:`__call__`.
        # Essentially the real wsgi application.
        self._stack = self._wsgi_inner
        for wrapper, args, kwargs in middleware:
            self._stack = wrapper(self._stack, *args, **kwargs)

    def _get_response(self, request):
        try:
            # TODO verify that request matches expected hostname, protocol and
            # mountpoint
            endpoint, kwargs = self._map_adapter.match(
                path_info=request.path, query_args=request.query_string,
            )

            binding = self._dispatcher.lookup(
                endpoint,
                method=request.environ.get('REQUEST_METHOD'),
                accept=request.environ.get('HTTP_ACCEPT'),
                accept_charset=request.environ.get('HTTP_ACCEPT_CHARSET'),
                accept_language=request.environ.get('HTTP_ACCEPT_LANGUAGE')
            )

            request.binding = binding

            return binding(self, request, **kwargs)
        except Exception:
            exc_type, exc_value, exc_traceback = sys.exc_info()

            handler = self._exception_dispatcher.lookup(
                exc_type,
                accept=request.environ.get('HTTP_ACCEPT'),
                accept_charset=request.environ.get('HTTP_ACCEPT_CHARSET'),
                accept_language=request.environ.get('HTTP_ACCEPT_LANGUAGE')
            )
            if handler is None:
                raise

            return handler(self, request, exc_type, exc_value, exc_traceback)

    def _wsgi_inner(self, wsgi_env, start_response):
        with self.request_class(wsgi_env) as request:
            with self._get_response(request) as response:
                yield from response(wsgi_env, start_response)

    def __call__(self, wsgi_env, start_response):
        wsgi_env['verktyg.application'] = self
        return self._stack(wsgi_env, start_response)

    def url_for(self, endpoint, **kwargs):
        """ construct the url corresponding to an endpoint name and parameters

        Unfortunately will only work if the application has been bound to a
        wsgi request.  If it is not then there is not generally enough
        information to construct full urls.  TODO.
        """
        return self._map_adapter.build(endpoint, values=kwargs)


class ApplicationBuilder(object):
    def __init__(
        self, *,
        default_redirect_handler=True, default_request_mixins=True
    ):
        self.config = {}

        self._application_bases = [BaseApplication]
        self._request_bases = [BaseRequest]

        self._middleware = []

        self._routes = []
        self._converters = {}
        self._bindings = []
        self._handlers = []

        if default_redirect_handler:
            self.add_exception_handlers(
                ExceptionHandler(RequestRedirect, _default_redirect_handler),
            )

        if default_request_mixins:
            self.add_request_mixins(
                requests.BaseRequest,
                requests.ETagRequestMixin,
                requests.AuthorizationMixin,
                requests.CommonRequestDescriptorsMixin,
            )

    def add_application_mixins(self, *mixins):
        for mixin in mixins:
            if mixin not in self._application_bases:
                self._application_bases.append(mixin)

    def add_request_mixins(self, *mixins):
        for mixin in mixins:
            if mixin not in self._request_bases:
                self._request_bases.append(mixin)

    def add_converters(self, **converters):
        self._converters.update(**converters)

    def add_routes(self, *routes):
        self._routes.extend(routes)

    def add_bindings(self, *views):
        self._bindings.extend(views)

    def expose(self, endpoint=None, *args, **kwargs):
        """ Decorator to bind a function to an endpoint and optionally the
        endpoint to a route.
        """
        def wrapper(f):
            nonlocal endpoint
            if endpoint is None:
                endpoint = f.__name__

            route = kwargs.pop('route', None)
            if route is not None:
                self.add_routes(Route(route, endpoint=endpoint))

            return expose(self, endpoint, *args, **kwargs)(f)
        return wrapper

    def add_middleware(self, middleware, *args, **kwargs):
        """ Wrap the application in a layer of wsgi middleware.

        :param middleware:
            a function which takes a wsgi application as it's first argument
            and returns a new wsgi application.  Any other args or kwargs are
            passed after.
        """
        self._middleware.append((middleware, args, kwargs))

    def add_exception_handlers(self, *handlers):
        self._handlers.extend(handlers)

    def add_exception_handler(self, exception_class, handler, **kwargs):
        """ Bind a function to render exceptions of the given class and all
        sub classes.

        Exception handlers take three arguments:
          * a reference to the application
          * a request object
          * the class of the exception
          * the exception instance
          * a traceback
        The last three arguments are the same as the return value of
        `sys.exc_info()`
        """
        self.add_exception_handlers(
            ExceptionHandler(exception_class, handler, **kwargs)
        )

    def exception_handler(self, exception_class, **kwargs):
        """ Decorator that can be used to bind an exception handler to the
        application.  Takes the same arguments as `ExceptionHandler`
        """
        def wrapper(handler):
            self.add_exception_handler(exception_class, handler, **kwargs)
            return handler
        return wrapper

    def __call__(self, app_root=''):
        class Application(*self._application_bases):
            pass

        class Request(*self._request_bases):
            pass

        return Application(
            app_root, self.config,
            routes=iter(self._routes),
            converters=dict(self._converters.items()),
            bindings=iter(self._bindings),
            handlers=iter(self._handlers),
            middleware=iter(self._middleware),
            request_class=Request,
        )



"""
    verktyg.testsuite
    ~~~~~~~~~~~~~~~~~

    :copyright:
        (c) 2015 Ben Mather, based on Werkzeug, see AUTHORS for more details.
    :license:
        BSD, see LICENSE for more details.

"""
import unittest

from verktyg.testsuite import (
    test_utils, test_datastructures, test_exceptions, test_http,
    test_accept_content_type, test_accept_language, test_accept_charset,
    test_accept, test_wsgi, test_requests, test_responses, test_routing,
    test_dispatch, test_views, test_application, test_urls,
)


loader = unittest.TestLoader()
suite = unittest.TestSuite((
    loader.loadTestsFromModule(test_utils),
    loader.loadTestsFromModule(test_datastructures),
    loader.loadTestsFromModule(test_exceptions),
    loader.loadTestsFromModule(test_http),
    loader.loadTestsFromModule(test_wsgi),
    loader.loadTestsFromModule(test_requests),
    loader.loadTestsFromModule(test_responses),
    loader.loadTestsFromModule(test_routing),
    loader.loadTestsFromModule(test_dispatch),
    loader.loadTestsFromModule(test_accept_content_type),
    loader.loadTestsFromModule(test_accept_language),
    loader.loadTestsFromModule(test_accept_charset),
    loader.loadTestsFromModule(test_accept),
    loader.loadTestsFromModule(test_views),
    loader.loadTestsFromModule(test_application),
    loader.loadTestsFromModule(test_urls),
))



"""
    verktyg
    ~~~~~~~

    :copyright: (c) 2014 by Ben Mather.
    :license: BSD, see LICENSE for more details.
"""
from verktyg.responses import BaseResponse, Response
from verktyg.routing import URLMap
from verktyg.dispatch import Dispatcher
from verktyg.application import ApplicationBuilder

__all__ = [
    'URLMap', 'Dispatcher', 'ApplicationBuilder', 'BaseResponse', 'Response',
]



"""
    verktyg.accept.content_type
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyright:
        (c) 2016 Ben Mather
    :license:
        BSD, see LICENSE for more details.
"""
import re

from verktyg.exceptions import NotAcceptable
from verktyg.accept import _base


_token_re_str = r'''
    (?:
        [a-zA-Z0-9]+
        (?:
            - [a-zA-Z0-9]+
        )*
    )
'''

# TODO
_value_re = re.compile(
    r'''
        ^
        {token} | \*
        $
    '''.format(token=_token_re_str), re.VERBOSE
)

# TODO
_content_type_value_re = re.compile(
    r'''
        ^
        (?: {token} | \*)
        /
        (?:
            (?:
                (?:{token} \.)?
                {token}
                (?: \+ {token})?
            ) | \*
        )
    '''.format(token=_token_re_str), re.VERBOSE
)


class _ContentTypeRange(_base.Range):
    def _validate_value(self, value):
        if _content_type_value_re.match(value) is None:
            raise ValueError("Invalid value: %r" % value)

    @property
    def type(self):
        type, _ = self.value.split('/')
        return type

    @property
    def subtype(self):
        _, subtype = self.value.split('/')
        return subtype


class ContentTypeAccept(_base.Accept):
    range_type = _ContentTypeRange


class ContentTypeAcceptability(_base.Acceptability):
    def __init__(
        self, content_type, *,
        type_matches, subtype_matches,
        q, qs=None
    ):
        super(ContentTypeAcceptability, self).__init__(
            content_type, match_quality=(
                type_matches, subtype_matches
            ),
            q=q, qs=qs
        )

    @property
    def content_type(self):
        return self._value

    @property
    def type_matches(self):
        return self._match_quality[0]

    @property
    def subtype_matches(self):
        return self._match_quality[1]

    @property
    def exact_match(self):
        return self._match_quality[0] and self._match_quality[1]


class ContentType(_base.Value):
    match_type = ContentTypeAcceptability

    @property
    def type(self):
        type, _ = self.value.split('/')
        return type

    @property
    def subtype(self):
        _, subtype = self.value.split('/')
        return subtype

    def _acceptability_for_option(self, option):
        if option.type == self.type:
            type_matches = True
        elif option.type == '*':
            type_matches = False
        else:
            raise NotAcceptable()

        if option.subtype == self.subtype:
            subtype_matches = True
        elif option.subtype == '*':
            subtype_matches = False
        else:
            raise NotAcceptable()

        return self.match_type(
            self, type_matches=type_matches, subtype_matches=subtype_matches,
            qs=self.qs, q=option.q
        )

    def to_header(self):
        """Returns a string suitable for use as a `Content-Type` header.
        """
        return "%s/%s" % (self.type, self.subtype)


def parse_accept_header(string):
    """Creates a new `ContentTypeAccept` object from an `Accept` header string.
    """
    return ContentTypeAccept(_base.split_accept_string(string))


def parse_content_type_header(string, qs=None):
    """Creates a new `ContentType` object from a mime type string.
    """
    return ContentType(string, qs=qs)



"""
    verktyg.http
    ~~~~~~~~~~~~

    :copyright:
        (c) 2015 Ben Mather, based on Werkzeug, see AUTHORS for more details.
    :license:
        BSD, see LICENSE for more details.
"""
from verktyg.http.basic import (
    Headers,
    ImmutableHeadersMixin,
    HeaderSet,
    FileStorage,
    wsgi_to_bytes,
    bytes_to_wsgi,
    parse_date,
    http_date,
    quote_header_value,
    unquote_header_value,
    unicodify_header_value,
    dump_options_header,
    dump_header,
    parse_list_header,
    parse_dict_header,
    parse_set_header,
    parse_options_header,
    remove_entity_headers,
    remove_hop_by_hop_headers,
    is_entity_header,
    is_hop_by_hop_header,
    is_byte_range_valid,
)
from verktyg.http.cache_control import (
    IfRange,
    Range,
    ContentRange,
    parse_if_range_header,
    parse_range_header,
    parse_content_range_header,
    RequestCacheControl,
    ResponseCacheControl,
    ETags,
    quote_etag,
    unquote_etag,
    parse_cache_control_header,
    parse_etags,
    generate_etag,
    is_resource_modified,
)
from verktyg.http.auth import (
    Authorization,
    WWWAuthenticate,
    parse_authorization_header,
    parse_www_authenticate_header,
)
from verktyg.http.cookies import (
    cookie_date,
    parse_cookie,
    dump_cookie,
)

from verktyg import exceptions
# reexport status codes.  Original definition is in in exceptions to avoid
# circular dependency
HTTP_STATUS_CODES = exceptions.HTTP_STATUS_CODES


__all__ = [
    'Headers',
    'ImmutableHeadersMixin',
    'HeaderSet',
    'FileStorage',
    'wsgi_to_bytes',
    'bytes_to_wsgi',
    'parse_date',
    'http_date',
    'quote_header_value',
    'unquote_header_value',
    'unicodify_header_value',
    'dump_options_header',
    'dump_header',
    'parse_list_header',
    'parse_dict_header',
    'parse_set_header',
    'parse_options_header',
    'remove_entity_headers',
    'remove_hop_by_hop_headers',
    'is_entity_header',
    'is_hop_by_hop_header',
    'is_byte_range_valid',
    'IfRange',
    'Range',
    'ContentRange',
    'parse_if_range_header',
    'parse_range_header',
    'parse_content_range_header',
    'RequestCacheControl',
    'ResponseCacheControl',
    'ETags',
    'quote_etag',
    'unquote_etag',
    'parse_cache_control_header',
    'parse_etags',
    'generate_etag',
    'is_resource_modified',
    'Authorization',
    'WWWAuthenticate',
    'parse_authorization_header',
    'parse_www_authenticate_header',
    'cookie_date',
    'parse_cookie',
    'dump_cookie',
]



"""
    verktyg.testsuite.test_accept_charset
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Tests for HTTP accept header parsing utilities.


    :copyright:
        (c) 2016 Ben Mather
    :license:
        BSD, see LICENSE for more details.
"""
import unittest

from verktyg.accept.charset import (
    parse_charset_header, parse_accept_charset_header,
    Charset, CharsetAccept,
)
from verktyg.exceptions import NotAcceptable


class CharsetTestCase(unittest.TestCase):
    def test_parse_accept_basic(self):
        accept = parse_accept_charset_header(
            'iso-8859-5'
        )

        range_ = next(iter(accept))
        self.assertEqual('iso-8859-5', range_.value)
        self.assertEqual(1, range_.q)

    def test_parse_accept_q(self):
        accept = parse_accept_charset_header(
            'ascii;q=0.5',
        )

        range_ = next(iter(accept))
        self.assertEqual('ascii', range_.value)
        self.assertEqual(0.5, range_.q)

    def test_parse_accept_params(self):
        with self.assertRaises(ValueError):
            parse_accept_charset_header(
                'utf-8;orange=black'
            )

    def test_parse_accept_multiple(self):
        accept = parse_accept_charset_header(
            'utf-8,'
            'ascii;q=0.5,'
            '*;q=0.1'
        )

        self.assertEqual(3, len(list(accept)))

    def test_parse(self):
        charset = parse_charset_header('utf-8')
        self.assertEqual('utf-8', charset.value)

    def test_serialize(self):
        charset = Charset('iso-8859-1', qs=0.5)
        self.assertEqual('iso-8859-1', charset.to_header())

    def test_serialize_accept(self):
        accept = CharsetAccept(['ascii'])
        self.assertEqual(accept.to_header(), 'ascii')

    def test_serialize_accept_with_q(self):
        accept = CharsetAccept([('utf-8', '0.5')])
        self.assertEqual(accept.to_header(), 'utf-8;q=0.5')

    def test_serialize_accept_redundant_q(self):
        accept = CharsetAccept([('utf-8', '1')])
        self.assertEqual(accept.to_header(), 'utf-8')

    def test_serialize_accept_multiple(self):
        accept = CharsetAccept([
            'utf-8',
            ('ascii', 0.5),
            ('*', 0.1),
        ])
        self.assertEqual(
            accept.to_header(),
            (
                'utf-8,'
                'ascii;q=0.5,'
                '*;q=0.1'
            )
        )

    def test_match_basic(self):
        accept = CharsetAccept(['utf-8'])

        acceptable = Charset('utf-8')
        unacceptable = Charset('latin-1')

        self.assertRaises(NotAcceptable, unacceptable.acceptability, accept)

        match = acceptable.acceptability(accept)
        self.assertEqual(acceptable, match.charset)
        self.assertTrue(match.exact_match)

    def test_match_wildcard(self):
        accept = CharsetAccept(['*'])

        charset = Charset('iso-8859-8')

        match = charset.acceptability(accept)
        self.assertEqual(charset, match.charset)
        self.assertFalse(match.exact_match)

    def test_match_quality(self):
        accept = CharsetAccept([('utf-8', '0.5')])

        no_qs = Charset('utf-8')
        qs = Charset('utf-8', qs=0.5)

        self.assertEqual(0.5, no_qs.acceptability(accept).quality)
        self.assertEqual(0.25, qs.acceptability(accept).quality)

    def test_match(self):
        accept = CharsetAccept([
            'utf-8',
            ('ascii', 0.5),
            ('*', 0.1),
        ])

        charset = Charset('utf-8')
        match = charset.acceptability(accept)
        self.assertEqual(1.0, match.quality)
        self.assertTrue(match.exact_match)

        charset = Charset('ascii')
        match = charset.acceptability(accept)
        self.assertEqual(0.5, match.quality)
        self.assertTrue(match.exact_match)

        charset = Charset('latin-1')
        match = charset.acceptability(accept)
        self.assertEqual(0.1, match.quality)
        self.assertFalse(match.exact_match)



"""
    verktyg.utils
    ~~~~~~~~~~~~~

    This module implements various utilities for WSGI applications.  Most of
    them are used by the request and response wrappers but especially for
    middleware development it makes sense to use them without the wrappers.

    :copyright:
        (c) 2015 Ben Mather, based on Werkzeug, see AUTHORS for more details.
    :license:
        BSD, see LICENSE for more details.
"""
import re
import os

from html import escape


_filename_ascii_strip_re = re.compile(r'[^A-Za-z0-9_.-]')
_windows_device_files = {
    'CON', 'AUX', 'COM1', 'COM2', 'COM3', 'COM4',
    'LPT1', 'LPT2', 'LPT3', 'PRN', 'NUL',
}
_missing = object()


class cached_property(property):
    """A decorator that converts a function into a lazy property.  The
    function wrapped is called the first time to retrieve the result
    and then that calculated result is used the next time you access
    the value::

        class Foo(object):

            @cached_property
            def foo(self):
                # calculate something important here
                return 42

    The class has to have a `__dict__` in order for this property to
    work.
    """
    # implementation detail: A subclass of python's builtin property
    # decorator, we override __get__ to check for a cached value. If one
    # choses to invoke __get__ by hand the property will still work as
    # expected because the lookup logic is replicated in __get__ for
    # manual invocation.

    def __init__(self, func, name=None, doc=None):
        self.__name__ = name or func.__name__
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self.func = func

    def __set__(self, obj, value):
        obj.__dict__[self.__name__] = value

    def __get__(self, obj, type=None):
        if obj is None:
            return self

        value = obj.__dict__.get(self.__name__, _missing)
        if value is _missing:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value


class header_property(object):
    """Decorator to create properties that wrap a request header.
    """

    def __init__(
        self, name, default=None, load_func=None, dump_func=None,
        read_only=False, doc=None,
    ):
        self.name = name
        self.default = default
        self.load_func = load_func
        self.dump_func = dump_func
        self.read_only = read_only
        self.__doc__ = doc

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        if self.name not in obj.headers:
            return self.default
        rv = obj.headers[self.name]
        if self.load_func is not None:
            try:
                rv = self.load_func(rv)
            except (ValueError, TypeError):
                rv = self.default
        return rv

    def __set__(self, obj, value):
        if self.read_only:
            raise AttributeError('read only property')
        if self.dump_func is not None:
            value = self.dump_func(value)
        obj.headers[self.name] = value

    def __delete__(self, obj):
        if self.read_only:
            raise AttributeError('read only property')
        obj.headers.pop(self.name, None)

    def __repr__(self):
        return '<%s %s>' % (
            self.__class__.__name__,
            self.name
        )


def get_content_type(mimetype, charset):
    """Returns the full content type string with charset for a mimetype.

    If the mimetype represents text the charset will be appended as charset
    parameter, otherwise the mimetype is returned unchanged.

    :param mimetype:
        The mimetype to be used as content type.
    :param charset:
        The charset to be appended in case it was a text mimetype.
    :return:
        The content type.
    """
    if (
        mimetype.startswith('text/') or
        mimetype == 'application/xml' or
        (mimetype.startswith('application/') and mimetype.endswith('+xml'))
    ):
        mimetype += '; charset=' + charset
    return mimetype


def secure_filename(filename):
    """Pass it a filename and it will return a secure version of it.  This
    filename can then safely be stored on a regular file system and passed
    to :func:`os.path.join`.  The filename returned is an ASCII only string
    for maximum portability.

    On windows systems the function also makes sure that the file is not
    named after one of the special device files.

    >>> secure_filename("My cool movie.mov")
    'My_cool_movie.mov'
    >>> secure_filename("../../../etc/passwd")
    'etc_passwd'
    >>> secure_filename(u'i contain cool \\xfcml\\xe4uts.txt')
    'i_contain_cool_umlauts.txt'

    The function might return an empty filename.  It's your responsibility
    to ensure that the filename is unique and that you generate random
    filename if the function returned an empty one.

    :param filename:
        The filename to secure
    """
    if isinstance(filename, str):
        from unicodedata import normalize
        filename = normalize('NFKD', filename).encode('ascii', 'ignore')
        filename = filename.decode('ascii')
    for sep in os.path.sep, os.path.altsep:
        if sep:
            filename = filename.replace(sep, ' ')
    filename = str(_filename_ascii_strip_re.sub('', '_'.join(
        filename.split()
    ))).strip('._')

    # on nt a couple of special files are present in each folder.  We
    # have to ensure that the target file is not such a filename.  In
    # this case we prepend an underline
    if (
        os.name == 'nt' and filename and
        filename.split('.')[0].upper() in _windows_device_files
    ):
        filename = '_' + filename

    return filename


def redirect(location, code=302, Response=None):
    """Returns a response object (a WSGI application) that, if called,
    redirects the client to the target location.  Supported codes are 301,
    302, 303, 305, and 307.  300 is not supported because it's not a real
    redirect and 304 because it's the answer for a request with a request
    with defined If-Modified-Since headers.

    :param location:
        The location the response should redirect to.
    :param code:
        The redirect status code. defaults to 302.
    :param class Response:
        A Response class to use when instantiating a response. The default is
        :class:`verktyg.responses.Response` if unspecified.
    """
    if Response is None:
        from verktyg.responses import Response

    display_location = escape(location)
    if isinstance(location, str):
        # Safe conversion is necessary here as we might redirect
        # to a broken URI scheme (for instance itms-services).
        from verktyg.urls import iri_to_uri
        location = iri_to_uri(location)
    response = Response(
        '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">\n'
        '<title>Redirecting...</title>\n'
        '<h1>Redirecting...</h1>\n'
        '<p>You should be redirected automatically to target URL: '
        '<a href="%s">%s</a>.  If not click the link.' %
        (escape(location), display_location),
        code, mimetype='text/html'
    )
    response.headers['Location'] = location
    return response


def append_slash_redirect(environ, code=301):
    """Redirects to the same URL but with a slash appended.  The behavior
    of this function is undefined if the path ends with a slash already.

    :param environ:
        The WSGI environment for the request that triggers the redirect.
    :param code:
        The status code for the redirect.
    """
    new_path = environ['PATH_INFO'].strip('/') + '/'
    query_string = environ.get('QUERY_STRING')
    if query_string:
        new_path += '?' + query_string
    return redirect(new_path, code)
